from typing import Annotated

from litestar import Controller, Request, get
from litestar.exceptions import NotFoundException
from litestar.plugins.pydantic import PydanticDTO
from sparkden.models.assistant import ContextOption
from sparkden.models.shared import BaseModel
from sparkden.shared.utils import dto_config


class ListContextOptionsResponse(BaseModel):
    context_options: list[ContextOption]


ListContextOptionsResponseDTO = PydanticDTO[
    Annotated[ListContextOptionsResponse, dto_config()]
]


class AssistantController(Controller):
    path = "/assistants"

    @get("/", sync_to_thread=True)
    def list_assistants(self) -> list[dict]:
        from sparkden.assistants import assistants

        return [
            assistant.model_dump(by_alias=True)
            for assistant in assistants.get_assistants(include_disabled=True)
        ]

    @get(
        "/{assistant_id:str}/context-creators/{context_creator_id:str}",
        return_dto=ListContextOptionsResponseDTO,
    )
    async def get_context_options(
        self,
        assistant_id: str,
        context_creator_id: str,
        request: Request,
        search_query: str | None = None,
    ) -> ListContextOptionsResponse:
        from sparkden.assistants import assistants

        assistant = assistants.get_assistant(assistant_id)
        if assistant is None:
            raise NotFoundException(f"Assistant {assistant_id} not found")

        context_creator = assistant.get_context_creator(context_creator_id)
        if context_creator is None:
            raise NotFoundException(f"Context creator {context_creator_id} not found")

        if context_creator.context_selector is None:
            raise NotFoundException(
                f"Context selector for {context_creator_id} not found"
            )

        context_options = await context_creator.context_selector.context_options(
            user_id=str(request.user.id), search_query=search_query
        )

        return ListContextOptionsResponse(context_options=context_options)



import redis as redis
import redis.asyncio as async_redis

from .utils import getenv

_async_redis_client: async_redis.Redis | None = None
_redis_client: redis.Redis | None = None


def get_redis_client() -> redis.Redis:
    global _redis_client
    if _redis_client is None:
        _redis_client = redis.Redis.from_url(
            getenv("REDIS_URL"), decode_responses=False
        )
    return _redis_client


def get_async_redis_client() -> async_redis.Redis:
    global _async_redis_client
    if _async_redis_client is None:
        _async_redis_client = async_redis.Redis.from_url(getenv("REDIS_URL"))
    return _async_redis_client



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sparkden.models.knowledge import DataChunk, ParseResult


class BaseDataSplitter(ABC):
    @abstractmethod
    async def split(self, data: "ParseResult") -> list["DataChunk"]:
        pass



import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import dramatiq

from .progress import JobProgressNotifier


class QueuedJobHandler(ABC):
    """Base class for handling queued jobs."""

    _job_worker: Any

    def __init__(self, job_type: str):
        self.job_type = job_type
        self.progress_notifier = JobProgressNotifier()

    @abstractmethod
    async def handle_job(self, job_id: str, user_id: str, **kwargs) -> Dict[str, Any]:
        """
        Handle the specific queued job.

        Args:
            job_id: Job ID
            user_id: User ID
            **kwargs: Job parameters

        Returns:
            Job result dict
        """
        pass

    async def notify_job_progress(
        self,
        job_id: str,
        progress_percent: int,
        message: str,
        user_id: str,
        database_id: Optional[str] = None,
        **extra_data,
    ):
        """Send notification about job progress."""
        await self.progress_notifier.notify(
            job_id=job_id,
            progress_percent=progress_percent,
            message=message,
            user_id=user_id,
            database_id=database_id,
            **extra_data,
        )


def queued_job(
    *,
    queue_name: str = "default",
    max_retries: int = 3,
    min_backoff: int = 15000,
    max_backoff: int = 300000,
    **dramatiq_options,
):
    """
    Decorator to create a queued job.

    Args:
        queue_name: Job queue name
        max_retries: Maximum number of retries
        min_backoff: Minimum backoff time (milliseconds)
        max_backoff: Maximum backoff time (milliseconds)
        **dramatiq_options: Other options for Dramatiq actor
    """

    def decorator(handler_class: type[QueuedJobHandler]):
        handler = handler_class(job_type=handler_class.__name__)

        @dramatiq.actor(
            queue_name=queue_name,
            max_retries=max_retries,
            min_backoff=min_backoff,
            max_backoff=max_backoff,
            actor_name=f"{handler.job_type}_worker",
            store_results=True,  # Enable result storage
            **dramatiq_options,
        )
        async def job_worker(
            job_id: Optional[str] = None, user_id: Optional[str] = None, **kwargs
        ):
            """Dramatiq job worker"""
            if job_id is None:
                job_id = str(uuid.uuid4())

            if not user_id:
                user_id = "anonymous"

            try:
                await handler.notify_job_progress(
                    job_id, 0, f"Job {handler.job_type} started", user_id
                )

                result = await handler.handle_job(job_id, user_id, **kwargs)

                await handler.notify_job_progress(
                    job_id, 100, f"Job {handler.job_type} completed", user_id
                )

                return result

            except Exception as e:
                await handler.notify_job_progress(
                    job_id, -1, f"Job {handler.job_type} failed: {str(e)}", user_id
                )
                raise

        # Use the job type as the function name to avoid conflicts
        job_worker.__name__ = f"{handler.job_type}_worker"

        # Attach the job worker to the handler class
        handler_class._job_worker = job_worker
        return handler_class

    return decorator



import os
from abc import ABC
from typing import Callable


class SparkdenWorker(ABC):
    def __init__(
        self,
        on_broker_ready: Callable,
    ):
        self.setup_environment()
        self.setup_broker()
        self.setup_queued_jobs()

        on_broker_ready()

    def setup_environment(self):
        """
        Load environment variables and configure API keys.
        """
        import dashscope
        from dotenv import find_dotenv, load_dotenv

        dotenv_path = find_dotenv(usecwd=True)
        load_dotenv(dotenv_path)
        dashscope.api_key = os.environ.get("DASHSCOPE_API_KEY")

    def setup_broker(self):
        """
        Setup broker.
        """
        import dramatiq

        from .broker import create_broker

        self.broker = create_broker()
        dramatiq.set_broker(self.broker)

    def setup_queued_jobs(self):
        """
        Setup job functions.
        """
        import sparkden.queued_job.data_source_job  # noqa: F401



import inspect
import logging
from typing import Any

from google.adk.tools import BaseTool, ToolContext
from google.genai import types
from pydantic import BaseModel
from typing_extensions import override

from sparkden.models.assistant import (
    ToolResponse,
    ToolResponseStatus,
    UserApprovalResult,
)

from ..helpers import validate_tool_param

logger = logging.getLogger(__name__)


class UserApprovalTool(BaseTool):
    item_name: str
    item_type: type[BaseModel]

    ITEM_PARAM_NAME = "item_to_approve"

    def __init__(self, item_name: str, item_type: type[BaseModel], description: str):
        self.item_name = item_name
        self.item_type = item_type
        super().__init__(
            name=f"request_{item_name}_approval",
            description=inspect.cleandoc(description),
            is_long_running=True,
        )

    @override
    def _get_declaration(self) -> types.FunctionDeclaration | None:
        # Get JSON schema for the item_type parameter
        item_schema = self.item_type.model_json_schema()

        # Create the parameters schema with the item_key as the property
        parameters_schema = types.JSONSchema.model_validate(
            {
                "type": "object",
                "properties": {self.ITEM_PARAM_NAME: item_schema},
                "required": [self.ITEM_PARAM_NAME],
                "additionalProperties": False,
            }
        )

        # Create the response schema for ToolResponse
        response_schema = types.JSONSchema.model_validate(
            {
                "type": "object",
                "properties": {
                    "result": UserApprovalResult.model_json_schema(),
                    "status": {
                        "type": "string",
                        "enum": ["pending", "success", "error"],
                        "description": "The status of the tool response",
                    },
                    "error": {
                        "type": "string",
                        "description": "The error message of the tool response",
                    },
                },
                "additionalProperties": False,
            }
        )

        return types.FunctionDeclaration(
            name=self.name,
            description=self.description,
            parameters=types.Schema.from_json_schema(json_schema=parameters_schema),
            response=types.Schema.from_json_schema(json_schema=response_schema),
        )

    @override
    async def run_async(
        self, *, args: dict[str, Any], tool_context: ToolContext
    ) -> ToolResponse:
        if self.ITEM_PARAM_NAME not in args:
            error_str = f"""Invoking `{self.name}()` failed as the following mandatory input parameters are not present:
{self.ITEM_PARAM_NAME}
You could retry calling this tool, but it is IMPORTANT for you to provide all the mandatory parameters."""
            return ToolResponse(
                status=ToolResponseStatus.ERROR,
                error=error_str,
            )

        validation_error = validate_tool_param(
            self.name, args[self.ITEM_PARAM_NAME], self.item_type
        )
        if validation_error:
            return validation_error

        tool_context.actions.skip_summarization = True
        return ToolResponse(
            result=None,
            status=ToolResponseStatus.PENDING,
        )



from enum import StrEnum
from uuid import uuid4

from google.adk.sessions.database_session_service import StorageSession
from litestar.plugins.sqlalchemy import base
from sqlalchemy import ForeignKey, ForeignKeyConstraint, Index
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column


class UserRole(StrEnum):
    ADMIN = "admin"
    USER = "user"


class DataSourceStatus(StrEnum):
    NOT_LOADED = "not_loaded"
    PROCESSING = "processing"
    READY = "ready"
    ERROR = "error"


class User(base.UUIDAuditBase):
    __tablename__ = "users"
    id: Mapped[str] = mapped_column(default=lambda: str(uuid4()), primary_key=True)
    name: Mapped[str]
    username: Mapped[str | None]
    password: Mapped[bytes | None]
    avatar: Mapped[str | None]
    role: Mapped[UserRole] = mapped_column(default=UserRole.USER)
    extra_info: Mapped[dict | None] = mapped_column(JSONB)


class ApiKey(base.UUIDAuditBase):
    __tablename__ = "api_keys"
    key_hash: Mapped[str] = mapped_column(unique=True)
    user_id: Mapped[str] = mapped_column(ForeignKey("users.id"))
    revoked: Mapped[bool] = mapped_column(default=False)


class KnowledgeDataSource(base.UUIDAuditBase):
    __tablename__ = "knowledge_data_sources"
    __table_args__ = (
        Index("idx_knowledge_data_sources_collection_id", "collection_id"),
    )

    name: Mapped[str]
    creator_id: Mapped[str]
    status: Mapped[DataSourceStatus]
    extra_info: Mapped[dict | None] = mapped_column(JSONB)
    collection_id: Mapped[str]


class Task(base.UUIDAuditBase):
    __tablename__ = "tasks"
    __table_args__ = (
        Index("idx_tasks_agent_session_id", "agent_session_id"),
        Index("idx_tasks_user_id_assistant_id", "user_id", "assistant_id"),
        ForeignKeyConstraint(
            ["assistant_id", "user_id", "agent_session_id"],
            [StorageSession.app_name, StorageSession.user_id, StorageSession.id],
            ondelete="CASCADE",
        ),
    )

    user_id: Mapped[str] = mapped_column(ForeignKey("users.id"))
    assistant_id: Mapped[str]
    agent_session_id: Mapped[str]
    title: Mapped[str]
    extra_info: Mapped[dict | None] = mapped_column(JSONB)



from datetime import datetime, timezone

import redis.asyncio as redis

from sparkden.shared.redis import get_async_redis_client

from .models import QueuedJob, QueuedJobStatus


class JobStatusService:
    _redis: redis.Redis
    _job_key_prefix = "queued_job:"
    _pubsub_channel_prefix = "queued_job_events:"

    def __init__(self, redis_instance: redis.Redis):
        self._redis = redis_instance

    def _get_job_key(self, job_id: str) -> str:
        return f"{self._job_key_prefix}{job_id}"

    def _get_pubsub_channel(self, job_id: str) -> str:
        return f"{self._pubsub_channel_prefix}{job_id}"

    async def get_job(self, job_id: str) -> QueuedJob | None:
        job_data = await self._redis.get(self._get_job_key(job_id))
        if not job_data:
            return None
        return QueuedJob.model_validate_json(job_data)

    async def create_job(self, job_id: str, user_id: str) -> QueuedJob:
        now = datetime.now(timezone.utc)
        job = QueuedJob(
            id=job_id,
            user_id=user_id,
            created_at=now,
            updated_at=now,
        )
        await self._update_and_publish(job)
        return job

    async def update_job_progress(
        self,
        job_id: str,
        progress: int,
        message: str | None = None,
        custom_status: str | None = None,
    ) -> QueuedJob:
        job = await self.get_job(job_id)
        if not job:
            raise ValueError(f"Job with id {job_id} not found")

        job.progress = progress
        if custom_status:
            job.custom_status = custom_status

        # In a real scenario, you might want to store the message somewhere,
        # but for now we'll just publish it. We'll add it to the model if needed.

        await self._update_and_publish(job)
        return job

    async def mark_job_as_running(
        self, job_id: str, custom_status: str | None = None
    ) -> QueuedJob:
        return await self._set_job_status(
            job_id, QueuedJobStatus.RUNNING, custom_status=custom_status
        )

    async def mark_job_as_success(self, job_id: str, result: dict) -> QueuedJob:
        return await self._set_job_status(
            job_id, QueuedJobStatus.SUCCESS, result=result
        )

    async def mark_job_as_failure(self, job_id: str, error: str) -> QueuedJob:
        return await self._set_job_status(job_id, QueuedJobStatus.FAILURE, error=error)

    async def _set_job_status(
        self,
        job_id: str,
        status: QueuedJobStatus,
        result: dict | None = None,
        error: str | None = None,
        custom_status: str | None = None,
    ) -> QueuedJob:
        job = await self.get_job(job_id)
        if not job:
            raise ValueError(f"Job with id {job_id} not found")

        job.status = status
        job.result = result
        job.error = error
        if custom_status:
            job.custom_status = custom_status

        if status in [QueuedJobStatus.SUCCESS, QueuedJobStatus.FAILURE]:
            job.progress = 100

        await self._update_and_publish(job)
        return job

    async def _update_and_publish(self, job: QueuedJob):
        job.updated_at = datetime.now(timezone.utc)
        job_json = job.model_dump_json(exclude_none=True, by_alias=True)
        await self._redis.set(self._get_job_key(job.id), job_json)
        await self._redis.publish(self._get_pubsub_channel(job.id), job_json)


_job_status_service: JobStatusService | None = None


async def get_job_status_service() -> JobStatusService:
    """
    Returns the singleton instance of the JobStatusService.
    """
    global _job_status_service
    if _job_status_service is None:
        redis_client = get_async_redis_client()
        _job_status_service = JobStatusService(redis_client)
    return _job_status_service



#!/usr/bin/env python3
"""
Dramatiq worker manager

Usage:
  python worker.py start [--dev]     # Start worker (dev mode)
  python worker.py stop              # Stop worker
  python worker.py status            # Check status
  python worker.py restart [--dev]   # Restart worker
  python worker.py test              # Test connections

Options:
  --dev        Development mode (file watching + verbose logs)
  --processes  Number of processes (default: 1)
  --threads    Number of threads (default: 1)
"""

import argparse
import os
import signal
import subprocess
import sys
import time
from pathlib import Path

from dotenv import load_dotenv

# env variables auto loaded in docker compose
if os.getenv("REDIS_URL") is None:
    load_dotenv()


class WorkerManager:
    def __init__(self):
        self.cwd = Path(os.getcwd())
        self.pid_file = self.cwd / "dramatiq.pid"

    def _setup_env(self):
        """Setup environment variables"""

        env = os.environ.copy()
        if not env.get("REDIS_URL"):
            env["REDIS_URL"] = "redis://localhost:6379"
        return env

    def _get_pid(self):
        """Get running PID"""
        if self.pid_file.exists():
            try:
                with open(self.pid_file, "r") as f:
                    pid = int(f.read().strip())
                # Check if process is still running
                try:
                    os.kill(pid, 0)
                    return pid
                except OSError:
                    # Process doesn't exist, remove pid file
                    self.pid_file.unlink()
                    return None
            except (ValueError, FileNotFoundError):
                return None
        return None

    def _save_pid(self, pid):
        """Save PID to file"""
        with open(self.pid_file, "w") as f:
            f.write(str(pid))

    def start(self, worker, queues, dev=False, processes=1, threads=1):
        """Start worker"""
        if self._get_pid():
            print("❌ Worker already running")
            return False

        env = self._setup_env()

        if not queues:
            print("❌ 请通过 --queues 指定要监听的队列名，如 queue1")
            return False

        cmd = [
            sys.executable,
            "-m",
            "dramatiq",
            "--processes",
            str(processes),
            "--threads",
            str(threads),
            "--queues",
            queues,
        ]

        if dev:
            cmd.extend(["--watch", str(self.cwd), "--verbose"])
        else:
            cmd.append("--verbose")

        # 使用传入的 worker 参数
        if not worker:
            print(
                "❌ 请通过 --worker 指定要加载的 worker 模块名，如 xxx.queued_tasks.handlers"
            )
            return False
        cmd.append(worker)

        print("🚀 Starting Dramatiq worker...")
        print(f"   Mode: {'Development' if dev else 'Production'}")
        print(f"   Processes: {processes}")
        print(f"   Threads: {threads}")
        print(f"   Redis URL: {env.get('REDIS_URL')}")
        print("=" * 50)

        try:
            # Start background process
            process = subprocess.Popen(
                cmd,
                env=env,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                preexec_fn=os.setsid if os.name != "nt" else None,
            )

            # Save PID
            self._save_pid(process.pid)

            print(f"✅ Worker started (PID: {process.pid})")

            if dev:
                print("📝 Development mode: Live logs (Ctrl+C to stop)")
                try:
                    if process.stdout is not None:
                        for line in process.stdout:
                            print(line.decode().strip())
                except KeyboardInterrupt:
                    print("\n🛑 Stopping log monitoring...")
                    self.stop()
            else:
                print("📝 Production mode: Logs")
                try:
                    if process.stdout is not None:
                        for line in process.stdout:
                            print(line.decode().strip())
                except KeyboardInterrupt:
                    print("\n🛑 Stopping worker...")

            return True

        except Exception as e:
            print(f"❌ Start failed: {e}")
            return False

    def stop(self):
        """Stop worker"""
        pid = self._get_pid()
        if not pid:
            print("ℹ️  Worker not running")
            return True

        try:
            print(f"🛑 Stopping worker (PID: {pid})...")

            try:
                if os.name == "nt":  # Windows
                    os.kill(pid, signal.SIGTERM)
                else:  # Unix/Linux/macOS
                    os.killpg(os.getpgid(pid), signal.SIGINT)
            except ProcessLookupError:
                # Process already gone, which is fine.
                pass

            # Wait for process to end
            for _ in range(50):
                try:
                    reaped_pid, _ = os.waitpid(pid, os.WNOHANG)
                    if reaped_pid == pid:
                        # Process has been successfully reaped.
                        break
                except ChildProcessError:
                    try:
                        os.kill(pid, 0)  # Check if the process exists
                    except OSError:
                        break  # Exit loop
                time.sleep(0.1)
            else:
                # Force kill
                print("⚠️  Process not responding, force stopping...")
                try:
                    if os.name == "nt":
                        os.kill(pid, signal.SIGKILL)
                    else:
                        os.killpg(os.getpgid(pid), signal.SIGKILL)
                except ProcessLookupError:
                    # Process already gone, which is fine.
                    pass

            # Remove PID file
            if self.pid_file.exists():
                self.pid_file.unlink()

            print("✅ Worker stopped")
            return True

        except Exception as e:
            print(f"❌ Stop failed: {e}")
            return False

    def status(self):
        """Check status"""
        pid = self._get_pid()
        if pid:
            print(f"✅ Worker running (PID: {pid})")
        else:
            print("❌ Worker not running")

        # Check Redis connection
        try:
            import redis

            redis_url = os.getenv("REDIS_URL", "redis://localhost:6379")
            r = redis.from_url(redis_url)
            r.ping()
            print(f"✅ Redis connected: {redis_url}")
        except Exception as e:
            print(f"❌ Redis connection failed: {e}")

    def restart(self, worker, queues, dev=False, processes=1, threads=1):
        """Restart worker"""
        print("🔄 Restarting worker...")
        self.stop()
        time.sleep(1)
        return self.start(worker, queues, dev, processes, threads)

    def test(self):
        """Test connections"""
        print("🧪 Testing connections...")

        # Setup environment
        env = self._setup_env()
        os.environ.update(env)

        try:
            # Test Redis connection
            import redis

            redis_url = env.get("REDIS_URL")
            r = redis.from_url(redis_url)
            r.ping()
            print(f"✅ Redis connection successful: {redis_url}")

            # Test Broker setup
            from sparkden.shared.queued_job.config import setup_job_queue_broker

            broker = setup_job_queue_broker()
            print(f"✅ Broker setup successful: {type(broker).__name__}")

            # Test Tasks import
            # from xxx import handlers  # noqa: F401

            print("✅ Tasks import successful")

            print("🎉 All tests passed!")
            return True

        except Exception as e:
            print(f"❌ Test failed: {e}")
            import traceback

            traceback.print_exc()
            return False


def main():
    manager = WorkerManager()

    def shutdown_handler(signum, frame):
        print(f"ℹ️  Signal {signum} received, shutting down...")
        manager.stop()
        sys.exit(0)

    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)

    parser = argparse.ArgumentParser(
        description="Dramatiq worker manager",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=__doc__,
    )

    parser.add_argument(
        "command",
        choices=["start", "stop", "status", "restart", "test"],
        help="Command to execute",
    )

    parser.add_argument(
        "--dev",
        action="store_true",
        help="Development mode (file watching + verbose logs)",
    )

    parser.add_argument(
        "--processes", type=int, default=1, help="Number of processes (default: 1)"
    )

    parser.add_argument(
        "--threads", type=int, default=1, help="Number of threads (default: 1)"
    )

    parser.add_argument(
        "--worker",
        type=str,
        help="Worker module to load (e.g. xxx.queued_tasks.handlers)",
        required=False,
    )

    parser.add_argument(
        "--queues",
        type=str,
        help="Comma-separated queue names (e.g. queue1)",
        required=True,
    )

    args = parser.parse_args()

    success = False
    if args.command == "start":
        success = manager.start(
            worker=args.worker,
            queues=args.queues,
            dev=args.dev,
            processes=args.processes,
            threads=args.threads,
        )
    elif args.command == "stop":
        success = manager.stop()
    elif args.command == "status":
        manager.status()
        success = True
    elif args.command == "restart":
        success = manager.restart(
            worker=args.worker,
            queues=args.queues,
            dev=args.dev,
            processes=args.processes,
            threads=args.threads,
        )
    elif args.command == "test":
        success = manager.test()

    sys.exit(0 if success else 1)


if __name__ == "__main__":
    main()



from sparkden.db.schema import DataSourceStatus
from sparkden.knowledge.data_loader import DataLoader
from sparkden.knowledge.registry import knowledge_collections
from sparkden.queued_job.decorator import JobContext, JobResult, queued_job
from sparkden.services.knowledge.data_source import KnowledgeDataSourceService


@queued_job(queue_name="knowledge")
async def data_source_job(
    ctx: JobContext,
    data_source_id: str,
    reload: bool = False,
) -> JobResult:
    """
    Asynchronously processes a data source, including loading, parsing,
    embedding, and indexing.
    """
    data_source_service = KnowledgeDataSourceService(user_id=ctx.user_id)

    try:
        await ctx.update_progress(5, custom_status="initializing")

        data_source = await data_source_service.get_data_source(data_source_id)
        if not data_source:
            raise ValueError(f"Data source {data_source_id} not found")

        collection = knowledge_collections.get_collection(data_source.collection_id)
        if not collection:
            raise ValueError(f"Collection {data_source.collection_id} not found")

        loader = DataLoader(
            data_source=data_source,
            vector_dimensions=collection.vector_dimensions,
            retrieval_mode=collection.retrieval_mode,
        )

        if reload:
            async for status in loader.reload():
                await ctx.update_progress(30, custom_status=status)
        else:
            async for status in loader.load():
                await ctx.update_progress(30, custom_status=status)

        await data_source_service.update_data_source(
            data_source_id, {"status": DataSourceStatus.READY}
        )
        await ctx.update_progress(100, custom_status="completed")

        return {"data_source_id": data_source_id}

    except Exception as e:
        # Update data source status to ERROR on failure
        await data_source_service.update_data_source(
            data_source_id, {"status": DataSourceStatus.ERROR}
        )
        # Re-raise the exception so the job is marked as failed
        raise e



from abc import ABC, abstractmethod
from typing import Literal, cast, overload

from sparkden.models.knowledge import RetrievalMode

DenseVector = list[float]
SparseVector = tuple[list[int], list[float]]
HybridVector = tuple[DenseVector, SparseVector]
DenseOutput = list[DenseVector]
SparseOutput = list[SparseVector]
HybridOutput = tuple[DenseOutput, SparseOutput]
TextType = Literal["document", "query"]

EmbeddingsOutput = DenseOutput | SparseOutput | HybridOutput
SingleEmbeddingOutput = DenseVector | SparseVector | HybridVector


class BaseEmbeddings(ABC):
    def __init__(
        self,
        *,
        model: str,
        dimensions: int,
    ):
        self.model = model
        self.dimensions = dimensions

    @overload
    def embed_documents(
        self,
        texts: list[str],
        retrieval_mode: Literal[RetrievalMode.DENSE],
    ) -> DenseOutput: ...
    @overload
    def embed_documents(
        self,
        texts: list[str],
        retrieval_mode: Literal[RetrievalMode.SPARSE],
    ) -> SparseOutput: ...
    @overload
    def embed_documents(
        self,
        texts: list[str],
        retrieval_mode: Literal[RetrievalMode.HYBRID],
    ) -> HybridOutput: ...

    def embed_documents(
        self,
        texts: list[str],
        retrieval_mode: RetrievalMode = RetrievalMode.DENSE,
    ) -> EmbeddingsOutput:
        return self._embed_texts(texts, "document", retrieval_mode)

    @overload
    def embed_query(
        self, text: str, retrieval_mode: Literal[RetrievalMode.DENSE]
    ) -> DenseVector: ...
    @overload
    def embed_query(
        self, text: str, retrieval_mode: Literal[RetrievalMode.SPARSE]
    ) -> SparseVector: ...
    @overload
    def embed_query(
        self, text: str, retrieval_mode: Literal[RetrievalMode.HYBRID]
    ) -> HybridVector: ...

    def embed_query(
        self,
        text: str,
        retrieval_mode: RetrievalMode = RetrievalMode.DENSE,
    ) -> SingleEmbeddingOutput:
        return self._embed_text(text, "query", retrieval_mode)

    def _embed_text(
        self, text: str, text_type: TextType, retrieval_mode: RetrievalMode
    ) -> SingleEmbeddingOutput:
        if retrieval_mode == RetrievalMode.DENSE:
            vectors = cast(
                DenseOutput, self._embed_texts([text], text_type, RetrievalMode.DENSE)
            )
            return vectors[0]
        elif retrieval_mode == RetrievalMode.SPARSE:
            sparse_vectors = cast(
                SparseOutput, self._embed_texts([text], text_type, RetrievalMode.SPARSE)
            )
            return sparse_vectors[0]
        elif retrieval_mode == RetrievalMode.HYBRID:
            vectors, sparse_vectors = cast(
                HybridOutput, self._embed_texts([text], text_type, RetrievalMode.HYBRID)
            )
            return vectors[0], sparse_vectors[0]
        else:
            raise ValueError(f"Invalid retrieval mode: {retrieval_mode}")

    @abstractmethod
    def _embed_texts(
        self,
        texts: list[str],
        text_type: TextType,
        retrieval_mode: RetrievalMode = RetrievalMode.DENSE,
    ) -> DenseOutput | SparseOutput | HybridOutput:
        pass



from typing import Any, cast

from sparkden.db.schema import DataSourceStatus
from sparkden.db.schema import KnowledgeDataSource as StorageDataSource
from sparkden.knowledge.data_loader import DataLoader
from sparkden.knowledge.registry import knowledge_collections
from sparkden.knowledge.vector_store import (
    FieldCondition,
    Filter,
    MatchValue,
    VectorStore,
)
from sparkden.models.knowledge import (
    DataChunk,
    DataSourceCreatorType,
    FileObject,
    KnowledgeDataSource,
    KnowledgeDataSourceCreate,
)
from sparkden.models.shared import OffsetPagination, OrderBy
from sparkden.services.base import BaseService
from sparkden.shared.minio import save_objects
from sqlalchemy import delete, func, select, update
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.ext.asyncio import AsyncSession


class KnowledgeDataSourceService(BaseService):
    async def list_data_sources(
        self, collection_id: str, *, limit: int = 30, offset: int = 0
    ) -> OffsetPagination[KnowledgeDataSource]:
        async with self.db_session_maker.begin() as db_session:
            query = (
                select(
                    StorageDataSource,
                    func.count().over().label("total_count"),
                )
                .where(StorageDataSource.collection_id == collection_id)
                .order_by(StorageDataSource.created_at.desc())
                .offset(offset)
                .limit(limit)
            )
            result = (await db_session.execute(query)).all()
            data_sources = [KnowledgeDataSource.from_storage(row[0]) for row in result]
            total = result[0][1] if result else 0
            return OffsetPagination(
                items=data_sources,
                total=total,
                limit=limit,
                offset=offset,
            )

    async def get_data_source(self, data_source_id: str) -> KnowledgeDataSource | None:
        async with self.db_session_maker.begin() as db_session:
            data_source = await db_session.get(StorageDataSource, data_source_id)
            if data_source:
                return KnowledgeDataSource.from_storage(data_source)
            return None

    async def list_data_source_chunks(
        self,
        collection_id: str,
        data_source_id: str,
        *,
        limit: int = 20,
        offset: int = 0,
    ) -> OffsetPagination[DataChunk]:
        filter = Filter(
            must=[
                FieldCondition(
                    key="data_source_id",
                    match=MatchValue(value=data_source_id),
                ),
            ]
        )

        vector_store = VectorStore(
            collection_name=collection_id,
        )

        chunks = await vector_store.filter_chunks(
            filter,
            limit,
            offset,
            OrderBy(field="sequence_in_data_source", desc=False),
        )

        items = [
            DataChunk(
                id=chunk.id,
                content=chunk.content,
                metadata=chunk.metadata,
            )
            for chunk in chunks
        ]

        return OffsetPagination(
            items=items,
            total=await vector_store.count_chunks(filter),
            limit=limit,
            offset=offset,
        )

    async def add_data_source(
        self, data_source: KnowledgeDataSourceCreate, auto_load: bool = True
    ) -> tuple[KnowledgeDataSource, bool, str | None]:
        """
        Adds a new data source.
        For file uploads, it dispatches an async job to process the data.

        Returns:
            A tuple of (KnowledgeDataSource, created_newly, job_id)
        """

        from sparkden.queued_job.data_source_job import data_source_job

        data_source_creator = knowledge_collections.get_data_source_creator(
            data_source.creator_id
        )
        if not data_source_creator:
            raise ValueError("Data source creator not found")

        file: FileObject | None = None
        if data_source_creator.type == DataSourceCreatorType.UPLOAD:
            file = data_source.pop_extra_info("file")
            if not file:
                raise ValueError("File not found")
            file = cast(FileObject, file)

        async with self.db_session_maker.begin() as db_session:
            # prevent duplicate file data source
            if file:
                existing_data_source = await get_data_source_by_file_hash(
                    data_source.collection_id, file.hash, db_session
                )
                if existing_data_source:
                    return existing_data_source, False, None

                data_source.set_extra_info("file_hash", file.hash)

            new_data_source = await create_data_source(data_source, db_session)

        if file:
            save_objects(
                bucket_name=new_data_source.collection_id,
                objects=[
                    file.model_copy(
                        update={"name": new_data_source.file_object_name},
                        deep=True,
                    )
                ],
            )

        if auto_load:
            # Dispatch async job after transaction commits
            message = data_source_job.send(
                data_source_id=new_data_source.id,
                user_id=self.user_id,
            )
            job_id = message.message_id
            updated_data_source = await self.update_data_source(
                new_data_source.id,
                {
                    "status": DataSourceStatus.PROCESSING,
                    "extra_info": {"job_id": job_id},
                },
            )
            return updated_data_source, True, job_id

        return new_data_source, True, None

    async def delete_data_source(self, data_source_id: str) -> None:
        async with self.db_session_maker.begin() as db_session:
            data_source = await db_session.scalar(
                delete(StorageDataSource)
                .where(
                    StorageDataSource.id == data_source_id,
                )
                .returning(StorageDataSource)
            )
        if not data_source:
            return

        loader = DataLoader(
            data_source=KnowledgeDataSource.from_storage(data_source),
        )
        await loader.unload()

    async def refresh_data_source(self, data_source_id: str) -> str:
        from sparkden.queued_job.data_source_job import data_source_job

        data_source = await self.get_data_source(data_source_id)
        if not data_source:
            raise ValueError("Data source not found")

        message = data_source_job.send(
            data_source_id=data_source.id,
            user_id=self.user_id,
            reload=True,
        )
        job_id = message.message_id
        await self.update_data_source(
            data_source.id,
            {"status": DataSourceStatus.PROCESSING, "extra_info": {"job_id": job_id}},
        )
        return job_id

    async def update_data_source(
        self,
        data_source_id: str,
        update_attributes: dict[str, Any],
    ) -> KnowledgeDataSource:
        """Update the status for a data source"""
        data_source = await self.get_data_source(data_source_id)
        if not data_source:
            raise ValueError("Data source not found")

        if update_attributes.get("extra_info"):
            update_attributes["extra_info"] = (
                data_source.extra_info or {}
            ) | update_attributes["extra_info"]

        async with self.db_session_maker.begin() as db_session:
            result = await db_session.scalar(
                update(StorageDataSource)
                .where(StorageDataSource.id == data_source_id)
                .values(update_attributes)
                .returning(StorageDataSource)
            )
            if not result:
                raise ValueError("Data source update failed")
            return KnowledgeDataSource.from_storage(result)


async def create_data_source(
    data_source: KnowledgeDataSourceCreate,
    db_session: AsyncSession,
) -> KnowledgeDataSource:
    data_source_attributes = {
        **data_source.model_dump(exclude_none=True),
    }
    data_source_id = await db_session.scalar(
        insert(StorageDataSource)
        .values(data_source_attributes)
        .on_conflict_do_update(
            index_elements=[StorageDataSource.id],
            set_=data_source_attributes,
        )
        .returning(StorageDataSource.id)
    )
    if not data_source_id:
        raise Exception("Failed to create data_source")
    if not data_source_id:
        raise Exception("Failed to create data_source")
    data_source_attributes["id"] = str(data_source_id)
    return KnowledgeDataSource(**data_source_attributes)


async def get_data_source_by_file_hash(
    collection_id: str,
    file_hash: str,
    db_session: AsyncSession,
) -> KnowledgeDataSource | None:
    query = select(StorageDataSource).where(
        StorageDataSource.collection_id == collection_id,
        StorageDataSource.extra_info["file_hash"].astext == file_hash,
    )
    data_source = await db_session.scalar(query)
    if data_source:
        return KnowledgeDataSource.from_storage(data_source)
    return None



from typing import Annotated, Any

from google.genai.types import Part
from litestar import Controller, Request, get, post
from litestar.datastructures import UploadFile
from litestar.di import Provide
from litestar.plugins.pydantic import PydanticDTO
from litestar.response import ServerSentEvent
from sparkden.models.assistant import ContextWithFile
from sparkden.models.shared import BaseModel, OffsetPagination
from sparkden.models.task import Task, TaskItem
from sparkden.services.task import TaskService
from sparkden.shared.utils import dto_config, snake_case_dict


class GetTaskResponse(BaseModel):
    task: Task | None


class CreateTaskParams(BaseModel):
    title: str
    assistant_id: str
    task_id: str | None = None


class SendMessageParams(BaseModel):
    assistant_id: str
    message: list[Part]
    state_delta: dict[str, Any] | None = None
    edit_message_id: str | None = None
    contexts: list[ContextWithFile] | None = None


CreateTaskParamsDTO = PydanticDTO[Annotated[CreateTaskParams, dto_config()]]
SendMessageParamsDTO = PydanticDTO[Annotated[SendMessageParams, dto_config()]]
TaskDTO = PydanticDTO[Annotated[Task, dto_config()]]
GetTaskResponseDTO = PydanticDTO[Annotated[GetTaskResponse, dto_config()]]
ListTasksResponseDTO = PydanticDTO[Annotated[OffsetPagination[TaskItem], dto_config()]]


def get_task_service(request: Request) -> TaskService:
    return TaskService(user_id=str(request.user.id))


class TaskController(Controller):
    path = "/tasks"
    dependencies = {
        "task_service": Provide(get_task_service, sync_to_thread=True),
    }
    request_max_body_size = 50 * 1024 * 1024  # 50MB

    @get("/", return_dto=ListTasksResponseDTO)
    async def list_tasks(
        self, task_service: TaskService, limit: int = 30, offset: int = 0
    ) -> OffsetPagination[TaskItem]:
        return await task_service.list_tasks(limit=limit, offset=offset)

    @post("/", dto=CreateTaskParamsDTO, return_dto=TaskDTO)
    async def create_task(
        self, task_service: TaskService, data: CreateTaskParams
    ) -> Task:
        return await task_service.create_task(
            title=data.title, assistant_id=data.assistant_id, task_id=data.task_id
        )

    @get("/{task_id:str}", return_dto=GetTaskResponseDTO)
    async def get_task(
        self, task_service: TaskService, task_id: str, assistant_id: str
    ) -> GetTaskResponse:
        task = await task_service.get_task(assistant_id=assistant_id, task_id=task_id)
        return GetTaskResponse(task=task)

    @post(
        "/{task_id:str}/run",
    )
    async def run_task(
        self,
        task_service: TaskService,
        task_id: str,
        request: Request,
    ) -> ServerSentEvent:
        form_data = await request.form()
        data = SendMessageParams.model_validate_json(form_data["payload"])

        if data.contexts:
            for context in data.contexts:
                file = form_data.get(context.id)
                if file and isinstance(file, UploadFile):
                    context.file = file

        return ServerSentEvent(
            task_service.run_task(
                assistant_id=data.assistant_id,
                task_id=task_id,
                message_content=data.message,
                state_delta=snake_case_dict(data.state_delta)
                if data.state_delta
                else None,
                edit_message_id=data.edit_message_id,
                contexts=data.contexts,
            )
        )



from google.adk.tools import ToolContext

from sparkden.models.assistant import ToolResponse


async def search_context(
    data_source_id: str, query: str, tool_context: ToolContext
) -> ToolResponse:
    """Retrieve information from context with data_source_id.

    Args:
        query: The query to search for.

    Returns:
        A list of data chunks.
    """
    from sparkden.assistants.helpers import assistant_context_collection_id
    from sparkden.db.schema import DataSourceStatus
    from sparkden.knowledge.vector_store.base import (
        FieldCondition,
        Filter,
        MatchValue,
    )
    from sparkden.models.assistant import ToolResponseStatus
    from sparkden.queued_job.search_context_job import search_context_job
    from sparkden.services.knowledge import KnowledgeService

    user_id = tool_context._invocation_context.user_id
    knowledge_service = KnowledgeService(user_id=user_id)

    try:
        data_source = await knowledge_service.get_data_source(data_source_id)
    except Exception:
        return ToolResponse(
            status=ToolResponseStatus.ERROR,
            error=f"Failed to get data source with id {data_source_id}",
        )

    if not data_source:
        return ToolResponse(
            status=ToolResponseStatus.ERROR,
            error=f"Data source {data_source_id} not found",
        )

    if data_source.status != DataSourceStatus.READY:
        tool_context.actions.skip_summarization = True

        message = search_context_job.send(
            data_source_id=data_source_id,
            query=query,
            user_id=user_id,
        )
        job_id = message.message_id

        return ToolResponse(
            result={
                "job_id": job_id,
                "data_source_creator_id": data_source.creator_id,
                "data_source_name": data_source.name,
            },
            status=ToolResponseStatus.PENDING,
        )

    filter = Filter(
        must=[
            FieldCondition(
                key="data_source_id",
                match=MatchValue(value=data_source_id),
            ),
        ],
    )
    data_chunks = await knowledge_service.search_collection(
        collection_id=assistant_context_collection_id(
            tool_context._invocation_context.app_name
        ),
        query=query,
        filter=filter,
    )

    return ToolResponse(
        result={
            "data_chunks": data_chunks,
            "data_source_creator_id": data_source.creator_id,
            "data_source_name": data_source.name,
        },
        status=ToolResponseStatus.SUCCESS,
    )



import hashlib
from enum import StrEnum
from functools import cached_property
from typing import Any, Sequence

from pydantic import Field
from sparkden.db.schema import (
    DataSourceStatus,
)
from sparkden.db.schema import KnowledgeDataSource as StorageDataSource
from sparkden.knowledge.data_parsers.base import (
    BaseDataParser,
)
from sparkden.knowledge.data_splitters.base import (
    BaseDataSplitter,
)
from sparkden.models.shared import BaseModel, ExtraInfoMixin


class VectorDistance(StrEnum):
    COSINE = "Cosine"
    EUCLID = "Euclid"
    DOT = "Dot"
    MANHATTAN = "Manhattan"


class RetrievalMode(StrEnum):
    DENSE = "dense"
    SPARSE = "sparse"
    HYBRID = "hybrid"


class DataSourceCreatorType(StrEnum):
    UPLOAD = "upload"
    CUSTOM = "custom"


class DataSourceCreator[DataType: BaseModel](ExtraInfoMixin, BaseModel):
    id: str
    type: DataSourceCreatorType = DataSourceCreatorType.UPLOAD
    data_parser: BaseDataParser[DataType] | None = Field(default=None, exclude=True)
    data_splitter: BaseDataSplitter | None = Field(default=None, exclude=True)


class KnowledgeCollection(BaseModel):
    id: str
    vector_dimensions: int = 1024
    vector_distance: VectorDistance = VectorDistance.COSINE
    retrieval_mode: RetrievalMode = RetrievalMode.HYBRID
    data_source_creators: Sequence[DataSourceCreator] | None = None


class KnowledgeDataSource(ExtraInfoMixin, BaseModel):
    id: str
    creator_id: str
    name: str
    collection_id: str
    status: DataSourceStatus

    @property
    def file_object_name(self) -> str:
        return f"{self.id}/original"

    @classmethod
    def from_storage(cls, data_source: StorageDataSource) -> "KnowledgeDataSource":
        return cls(
            id=str(data_source.id),
            creator_id=data_source.creator_id,
            name=data_source.name,
            status=data_source.status,
            extra_info=data_source.extra_info,
            collection_id=data_source.collection_id,
        )


class KnowledgeDataSourceCreate(KnowledgeDataSource):
    id: str | None = None
    status: DataSourceStatus = DataSourceStatus.NOT_LOADED


class FileObject(BaseModel):
    name: str
    content: bytes
    content_type: str

    @cached_property
    def hash(self) -> str:
        return hashlib.sha256(self.content).hexdigest()


class ParseResultType(StrEnum):
    TEXT = "text"
    MARKDOWN = "markdown"
    TABLE = "table"


class FileExtractConfig(BaseModel):
    object_base_name: str
    parsed_object_name: str
    bucket_name: str


class ParseResult(BaseModel):
    data: str | list
    data_type: ParseResultType
    extracted_files: list[FileObject] = []


class DataChunk(BaseModel):
    id: str
    content: str
    metadata: dict[str, Any]


class ScoredDataChunk(DataChunk):
    score: float
    retrieval_mode: RetrievalMode



from typing import Any

import pymupdf
from sparkden.models.knowledge import (
    FileExtractConfig,
    FileObject,
    ParseResult,
    ParseResultType,
)
from sparkden.models.shared import BaseModel
from sparkden.shared.pymupdf_rag import to_markdown

from .base import BaseDataParser


class FileData(BaseModel):
    file: FileObject


class FileParser(BaseDataParser):
    def __init__(self):
        super().__init__(FileData)

    async def parse(
        self,
        input_data: dict[str, Any] | FileData,
        file_extract_config: FileExtractConfig,
    ) -> ParseResult:
        data = self._parse_input_data(input_data)

        if data.file.content_type == "text/plain":
            return self._parse_text_file(data.file, file_extract_config)
        elif data.file.content_type == "application/pdf":
            return self._parse_pdf_file(data.file, file_extract_config)
        else:
            raise ValueError(f"Unsupported file type: {data.file.content_type}")

    @staticmethod
    def _parse_text_file(
        file: FileObject, file_extract_config: FileExtractConfig
    ) -> ParseResult:
        return ParseResult(
            data=file.content.decode("utf-8"),
            data_type=ParseResultType.TEXT,
            extracted_files=[
                FileObject(
                    name=file_extract_config.parsed_object_name,
                    content=file.content,
                    content_type=file.content_type,
                ),
            ],
        )

    @staticmethod
    def _parse_pdf_file(
        file: FileObject, file_extract_config: FileExtractConfig
    ) -> ParseResult:
        doc = pymupdf.open(stream=file.content, filetype="pdf")
        extracted_images: list[FileObject] = []

        def extract_image(image_data: bytes, image_name: str) -> str:
            extracted_images.append(
                FileObject(
                    name=image_name,
                    content=image_data,
                    content_type="image/png",
                )
            )
            return f"/images/{file_extract_config.bucket_name}/{image_name}"

        markdown = to_markdown(
            doc,
            filename=file_extract_config.object_base_name,
            write_images=extract_image is not None,
            on_write_image=extract_image,
            image_format="png",
        )

        return ParseResult(
            data=markdown,
            data_type=ParseResultType.MARKDOWN,
            extracted_files=[
                FileObject(
                    name=file_extract_config.parsed_object_name,
                    content=markdown.encode("utf-8"),
                    content_type="text/markdown",
                ),
                *extracted_images,
            ],
        )



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sparkden.models.assistant import ContextOption


class ContextSelector(ABC):
    @abstractmethod
    async def context_options(
        self, user_id: str, search_query: str | None = None
    ) -> "list[ContextOption]":
        pass



from sparkden.db.schema import User as StorageUser
from sparkden.db.schema import UserRole
from sparkden.models.shared import BaseModel
from sparkden.shared.utils import getenv

ADMIN_USERS = [username.strip() for username in getenv("ADMIN_USERS", "").split(",")]


class User(BaseModel):
    id: str
    name: str
    username: str | None
    avatar: str | None
    role: UserRole
    extra_info: dict | None

    @classmethod
    def from_storage(cls, user: StorageUser) -> "User":
        return cls(
            id=user.id,
            name=user.name,
            username=user.username,
            avatar=user.avatar,
            role=user.role,
            extra_info=user.extra_info,
        )

    @property
    def is_admin(self) -> bool:
        return self.role == UserRole.ADMIN or self.username in ADMIN_USERS



from litestar import get
from litestar.types import Scope
from litestar.utils.scope.state import ScopeState


@get("/csrf-token", exclude_from_auth=True)
async def csrf_token(scope: Scope) -> str | None:
    token = ScopeState.from_scope(scope).csrf_token
    if not isinstance(token, str):
        return None
    return token



from datetime import timedelta
from os.path import basename

from litestar import Controller, get
from litestar.exceptions import NotFoundException
from litestar.response import Redirect
from sparkden.shared.minio import get_minio_client, object_exists


class FilesController(Controller):
    @get("/images/{bucket_name:str}/{object_name:path}")
    async def images(
        self, bucket_name: str, object_name: str, download: bool = False
    ) -> Redirect:
        if not object_exists(bucket_name, object_name):
            raise NotFoundException(
                f"File not found: bucket={bucket_name}, object_name={object_name}"
            )

        minio_client = get_minio_client()
        disposition_type = "attachment" if download else "inline"
        disposition = f'{disposition_type}; filename="{basename(object_name)}"'
        presigned_url = minio_client.presigned_get_object(
            bucket_name,
            object_name,
            expires=timedelta(seconds=60),
            response_headers={
                "response-content-disposition": disposition,
            },
        )

        return Redirect(
            path=presigned_url,
            status_code=303,
        )



from .base import (
    Distance,
    Field,
    FieldCondition,
    Filter,
    MatchValue,
    SparseVectorParams,
    VectorParams,
)
from .qdrant import QdrantVectorStore

# TODO: determine which vector store to use by app config
VectorStore = QdrantVectorStore

__all__ = [
    "VectorStore",
    "Field",
    "Filter",
    "MatchValue",
    "FieldCondition",
    "Distance",
    "VectorParams",
    "SparseVectorParams",
]



import hashlib

from bcrypt import gensalt, hashpw
from google.adk.sessions.database_session_service import Base as ADKBase
from sparkden.shared.pg import get_engine, get_session_maker
from sparkden.shared.utils import getenv
from sqlalchemy import exists, insert, select
from sqlalchemy.exc import OperationalError

from .schema import ApiKey, User, UserRole, base


async def init_db() -> None:
    await seed_db()


async def seed_db() -> None:
    async with get_engine().begin() as conn:
        try:
            await conn.run_sync(ADKBase.metadata.create_all)
            await conn.run_sync(base.UUIDAuditBase.metadata.create_all)
        except OperationalError as exc:
            print(f"Could not create target metadata.  Reason: {exc}")

    async with get_session_maker().begin() as db_session:
        user_exists = await db_session.scalar(
            select(exists().where(User.id.is_not(None)))
        )
        if not user_exists:
            try:
                api_keys = getenv("API_KEYS", "").split(",")
                users_data = [
                    {
                        "name": user,
                        "username": f"test{index + 1}",
                        "password": hashpw("password123".encode(), gensalt()),
                        "extra_info": {"college": "计算机学院"},
                        "role": UserRole.ADMIN if index in [0, 1] else UserRole.USER,
                    }
                    for index, user in enumerate(users)
                ]
                user_ids = await db_session.scalars(
                    insert(User).values(users_data).returning(User.id)
                )

                if len(api_keys) > 0:
                    api_keys_data = [
                        {
                            "key_hash": hashlib.sha256(api_key.encode()).hexdigest(),
                            "user_id": user_id,
                        }
                        for user_id, api_key in zip(user_ids, api_keys)
                    ]
                    await db_session.execute(insert(ApiKey).values(api_keys_data))

                print(f"Successfully seeded {len(users)} users")
            except Exception as exc:
                print(f"Error seeding users: {exc}")
                raise


users = [
    "张三",
    "李四",
    "王五",
    "赵六",
    "孙七",
    "周八",
    "吴九",
    "郑十",
    "John Smith",
    "Alice Brown",
]



from abc import ABC, abstractmethod
from enum import StrEnum
from itertools import islice
from typing import Any, Generator, Literal

from qdrant_client import models

from sparkden.models.knowledge import DataChunk, RetrievalMode, ScoredDataChunk
from sparkden.models.shared import BaseModel, OrderBy

from ..embeddings.base import BaseEmbeddings, DenseOutput, SparseOutput

Distance = models.Distance
VectorParams = models.VectorParams
SparseVectorParams = models.SparseVectorParams
Filter = models.Filter
MatchValue = models.MatchValue
MatchAny = models.MatchAny
FieldCondition = models.FieldCondition
SparseVector = models.SparseVector


class Field(BaseModel):
    name: str
    type: Literal[
        "string", "integer", "float", "bool", "text", "datetime", "uuid", "geo"
    ]
    default: Any = None


class SearchMode(StrEnum):
    DENSE = "dense"
    SPARSE = "sparse"
    HYBRID = "hybrid"


class BaseVectorStore(ABC):
    VECTOR_NAME = "vector"
    SPARSE_VECTOR_NAME = "sparse_vector"
    CONTENT_PAYLOAD_KEY = "content"

    def __init__(
        self,
        collection_name: str,
        retrieval_mode: RetrievalMode = RetrievalMode.DENSE,
        hybrid_dense_ratio: float = 0.6,
    ):
        self.collection_name = collection_name
        self.retrieval_mode = retrieval_mode
        self.hybrid_dense_ratio = hybrid_dense_ratio

    @classmethod
    @abstractmethod
    async def create_collection(
        cls,
        collection_name: str,
        vector_name: str | None = None,
        vector_params: VectorParams | None = None,
        sparse_vector_name: str | None = None,
        sparse_vector_params: SparseVectorParams | None = None,
        payload_fields: list[Field] | None = None,
        recreate_if_exists: bool = False,
    ) -> None:
        pass

    @classmethod
    @abstractmethod
    async def create_index(cls, collection_name: str, field: Field) -> None:
        pass

    async def add_chunks(
        self,
        chunks: list[DataChunk],
        *,
        embeddings: BaseEmbeddings,
        batch_size: int = 10,
    ) -> list[str]:
        added_ids = []
        for vectors, sparse_vectors, payloads, ids in self._generate_batches(
            chunks, embeddings, batch_size
        ):
            await self.add_vectors(vectors, sparse_vectors, payloads, ids)
            added_ids.extend(ids)

        return added_ids

    def _generate_batches(
        self,
        chunks: list[DataChunk],
        embeddings: BaseEmbeddings,
        batch_size: int = 10,
    ) -> Generator[
        tuple[DenseOutput | None, SparseOutput | None, list[dict], list[str]],
        None,
        None,
    ]:
        texts = iter([chunk.content for chunk in chunks])
        metadatas = iter([chunk.metadata for chunk in chunks])
        ids = iter([chunk.id for chunk in chunks])

        while batch_texts := list(islice(texts, batch_size)):
            batch_metadatas = list(islice(metadatas, batch_size))
            batch_ids = list(islice(ids, batch_size))

            vectors = None
            sparse_vectors = None
            if self.retrieval_mode == RetrievalMode.DENSE:
                vectors = embeddings.embed_documents(batch_texts, self.retrieval_mode)
            elif self.retrieval_mode == RetrievalMode.SPARSE:
                sparse_vectors = embeddings.embed_documents(
                    batch_texts, self.retrieval_mode
                )
            elif self.retrieval_mode == RetrievalMode.HYBRID:
                vectors, sparse_vectors = embeddings.embed_documents(
                    batch_texts, self.retrieval_mode
                )

            payloads = [
                {**metadata, self.CONTENT_PAYLOAD_KEY: text}
                for text, metadata in zip(batch_texts, batch_metadatas)
            ]

            yield vectors, sparse_vectors, payloads, batch_ids

    @abstractmethod
    async def add_vectors(
        self,
        vectors: DenseOutput | None,
        sparse_vectors: SparseOutput | None,
        payloads: list[dict],
        ids: list[str] | None = None,
    ) -> None:
        pass

    @abstractmethod
    async def delete_chunks(self, chunk_selector: list[str] | Filter) -> None:
        pass

    @abstractmethod
    async def count_chunks(self, filter: Filter) -> int:
        pass

    @abstractmethod
    async def filter_chunks(
        self,
        filter: Filter,
        limit: int = 20,
        offset: int = 0,
        order_by: OrderBy | None = None,
    ) -> list[ScoredDataChunk]:
        pass



from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)

from .utils import getenv

# 不立即创建 Engine，而是提供创建函数
_engine: AsyncEngine | None = None
_session_maker: async_sessionmaker[AsyncSession] | None = None


def get_engine() -> AsyncEngine:
    DATABASE_URL = getenv("DATABASE_URL")
    global _engine
    if _engine is None:
        _engine = create_async_engine(DATABASE_URL)
    return _engine


def get_session_maker() -> async_sessionmaker[AsyncSession]:
    global _session_maker
    if _session_maker is None:
        _session_maker = async_sessionmaker(get_engine(), expire_on_commit=False)
    return _session_maker


async def dispose_engine():
    global _engine
    if _engine is not None:
        await _engine.dispose()
        _engine = None



from uuid import uuid4

from langchain_text_splitters import (
    MarkdownHeaderTextSplitter,
    RecursiveCharacterTextSplitter,
)
from sparkden.models.knowledge import DataChunk, ParseResult, ParseResultType

from .base import BaseDataSplitter


class DefaultSplitter(BaseDataSplitter):
    async def split(self, data: ParseResult) -> list[DataChunk]:
        if data.data_type == ParseResultType.MARKDOWN:
            assert isinstance(data.data, str)
            return self._split_markdown(data.data)
        elif data.data_type == ParseResultType.TEXT:
            assert isinstance(data.data, str)
            return self._split_text(data.data)
        else:
            raise ValueError(f"Unsupported data type: {data.data_type}")

    def _split_markdown(self, markdown: str) -> list[DataChunk]:
        markdown_splitter = MarkdownHeaderTextSplitter(
            headers_to_split_on=[
                ("#", "Header1"),
                ("##", "Header2"),
                ("###", "Header3"),
                ("####", "Header4"),
                ("#####", "Header5"),
                ("######", "Header6"),
            ],
        )
        chunks = markdown_splitter.split_text(markdown)
        recursive_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1000, chunk_overlap=200
        )
        chunks = recursive_splitter.split_documents(chunks)
        return [
            DataChunk(
                id=str(uuid4().hex),
                content=chunk.page_content,
                metadata=chunk.metadata,
            )
            for chunk in chunks
        ]

    def _split_text(self, text: str) -> list[DataChunk]:
        recursive_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1000, chunk_overlap=200
        )

        chunks = recursive_splitter.create_documents([text])

        return [
            DataChunk(
                id=str(uuid4().hex),
                content=chunk.page_content,
                metadata=chunk.metadata,
            )
            for chunk in chunks
        ]



import asyncio
import json
import uuid
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Optional, Set

from .config import create_redis_async_client


class JobProgressNotifier:
    """Job progress notifier with caching support."""

    def __init__(self):
        self.redis_client = create_redis_async_client()
        self.cache_ttl = 3600  # Cache for 1 hour after completion

    async def _cache_job_state(
        self, job_id: str, event_data: dict, is_completed: bool = False
    ) -> None:
        """Cache the latest job state in Redis.

        Args:
            job_id: Job ID
            event_data: Event data to cache
            is_completed: Whether this is a completion state
        """
        try:
            cache_key = f"job_state:{job_id}"

            # Add caching metadata
            cache_data = {
                **event_data,
                "cached_at": asyncio.get_event_loop().time(),
                "is_completed": is_completed,
            }

            # Set TTL based on completion status
            ttl = self.cache_ttl if is_completed else None

            await self.redis_client.set(cache_key, json.dumps(cache_data), ex=ttl)

        except Exception as e:
            print(f"Warning: Failed to cache job state for {job_id}: {e}")

    async def notify(
        self,
        job_id: str,
        progress_percent: int,
        message: str,
        user_id: str,
        database_id: Optional[str] = None,
        **extra_data: Any,
    ) -> None:
        """
        Send job progress notification to Redis Pub/Sub and cache the state.

        Args:
            job_id: Job ID
            progress_percent: Progress percentage (0-100, negative for error)
            message: Progress message
            user_id: User ID
            database_id: Database ID (optional)
            **extra_data: Extra data
        """
        if progress_percent < 0:
            status = "error"
        elif progress_percent >= 100:
            status = "completed"
        else:
            status = "running"

        event_data = {
            "job_id": job_id,
            "progress_percent": progress_percent,
            "message": message,
            "status": status,
            "user_id": user_id,
            "timestamp": str(uuid.uuid4()),
            **extra_data,
        }

        if database_id:
            event_data["database_id"] = database_id

        # Cache the job state for future retrieval
        is_completed = status in ["completed", "error"]
        await self._cache_job_state(job_id, event_data, is_completed)

        # Publish to Redis pub/sub channel
        channel = f"job_progress:{job_id}"
        await self.redis_client.publish(channel, json.dumps(event_data))

        # Logging
        print(
            f"[User: {user_id}] Job {job_id}"
            + (f" (DB: {database_id})" if database_id else "")
            + f": Progress {progress_percent}% - {message}"
        )

    async def get_cached_job_state(self, job_id: str) -> Optional[dict]:
        """Get cached job state from Redis.

        Args:
            job_id: Job ID to retrieve state for

        Returns:
            Cached job state dict or None if not found
        """
        try:
            cache_key = f"job_state:{job_id}"
            cached_data = await self.redis_client.get(cache_key)

            if cached_data:
                return json.loads(cached_data)

            return None

        except Exception as e:
            print(f"Warning: Failed to retrieve cached job state for {job_id}: {e}")
            return None


class QueueJobStreamService:
    """Queue job stream service for managing job progress streams with caching support."""

    def __init__(self):
        self.redis_client = create_redis_async_client()
        self._active_streams: Set[str] = set()

    def get_active_streams(self) -> Set[str]:
        """Get currently active streams."""
        return self._active_streams.copy()

    @asynccontextmanager
    async def stream_manager(self, job_id: str):
        """Context manager for managing job progress streams."""
        self._active_streams.add(job_id)
        try:
            yield
        finally:
            self._active_streams.discard(job_id)

    async def get_cached_job_state(self, job_id: str) -> Optional[dict]:
        """Get cached job state from Redis.

        Args:
            job_id: Job ID to retrieve state for

        Returns:
            Cached job state dict or None if not found
        """
        try:
            cache_key = f"job_state:{job_id}"
            cached_data = await self.redis_client.get(cache_key)

            if cached_data:
                return json.loads(cached_data)

            return None

        except Exception as e:
            print(f"Warning: Failed to retrieve cached job state for {job_id}: {e}")
            return None

    async def stream_queued_job_progress(
        self, job_id: str
    ) -> AsyncGenerator[str, None]:
        """
        Stream job progress updates as Server-Sent Events (SSE).

        Args:
            job_id: Job ID to monitor

        Yields:
            SSE formatted strings with job progress updates
        """
        async with self.stream_manager(job_id):
            try:
                yield self._format_sse_event(
                    "connected",
                    {
                        "job_id": job_id,
                        "message": "Connected to queued job progress stream",
                    },
                )

                # Subscribe to Redis channel
                channel = f"job_progress:{job_id}"
                pubsub = self.redis_client.pubsub()
                await pubsub.subscribe(channel)

                try:
                    timeout = 300  # 5 minutes
                    start_time = asyncio.get_event_loop().time()

                    async for message in pubsub.listen():
                        if message["type"] == "message":
                            try:
                                data = json.loads(message["data"])

                                yield self._format_sse_event("progress", data)

                                if data.get("status") in ["completed", "error"]:
                                    yield self._format_sse_event(
                                        "finished",
                                        {
                                            "job_id": job_id,
                                            "final_status": data.get("status"),
                                        },
                                    )
                                    break

                            except json.JSONDecodeError:
                                continue

                        # Handle timeout
                        if asyncio.get_event_loop().time() - start_time > timeout:
                            yield self._format_sse_event(
                                "timeout", {"message": "Stream timeout after 5 minutes"}
                            )
                            break

                finally:
                    await pubsub.unsubscribe(channel)
                    await pubsub.aclose()

            except GeneratorExit:
                # TODO
                pass

            except Exception as e:
                yield self._format_sse_event(
                    "error", {"job_id": job_id, "error": str(e)}
                )

            finally:
                try:
                    yield self._format_sse_event(
                        "closed", {"job_id": job_id, "message": "Stream closed"}
                    )
                except GeneratorExit:
                    pass

    def _format_sse_event(self, event_type: str, data: dict) -> str:
        """Format SSE event string."""
        return f"event: {event_type}\ndata: {json.dumps(data)}\n\n"


job_progress_notifier = JobProgressNotifier()
queue_job_stream_service = QueueJobStreamService()


async def notify_job_progress(
    job_id: str,
    progress_percent: int,
    message: str,
    user_id: str,
    database_id: Optional[str] = None,
    **extra_data: Any,
) -> None:
    """
    Global function to send job progress notifications.

    Args:
        job_id: Job ID
        progress_percent: Progress percentage (0-100, negative for error)
        message: Progress message
        user_id: User ID
        database_id: Database ID (optional)
        **extra_data: Extra data
    """
    await job_progress_notifier.notify(
        job_id=job_id,
        progress_percent=progress_percent,
        message=message,
        user_id=user_id,
        database_id=database_id,
        **extra_data,
    )



import uuid
from datetime import datetime
from typing import Any, Literal

from google.adk.events import Event
from google.adk.sessions import Session
from google.genai.types import FunctionCall, FunctionResponse
from sparkden.db.schema import Task as StorageTask
from sparkden.models.shared import BaseModel, ExtraInfoMixin
from sparkden.shared.utils import camelize_dict


class ContentPart(BaseModel):
    id: str
    thought: bool | None = None
    function_call: FunctionCall | None = None
    function_response: FunctionResponse | None = None
    text: str | None = None


class TaskMessage(BaseModel):
    id: str
    type: Literal["user", "assistant"]
    content: list[ContentPart]
    created_at: datetime
    partial: bool | None = None

    @classmethod
    def from_event(cls, event: Event) -> "TaskMessage":
        content_parts = (
            event.content.parts if event.content and event.content.parts else []
        )
        return TaskMessage(
            id=event.id,
            type="user" if event.author == "user" else "assistant",
            created_at=datetime.fromtimestamp(event.timestamp),
            content=[
                ContentPart(
                    id=str(uuid.uuid4()),
                    thought=part.thought,
                    function_call=part.function_call,
                    function_response=part.function_response,
                    text=part.text,
                )
                for part in content_parts
            ],
            partial=event.partial,
        )


class TaskItem(BaseModel):
    id: str
    assistant_id: str
    title: str
    updated_at: datetime

    @classmethod
    def from_storage(cls, task: "StorageTask") -> "TaskItem":
        return cls(
            id=str(task.id),
            assistant_id=task.assistant_id,
            title=task.title,
            updated_at=task.updated_at,
        )


class Task(ExtraInfoMixin, BaseModel):
    id: str
    assistant_id: str
    title: str
    messages: list[TaskMessage]
    state: dict[str, Any]
    updated_at: datetime

    @classmethod
    def from_storage(cls, task: StorageTask, session: Session) -> "Task":
        return cls(
            id=str(task.id),
            assistant_id=task.assistant_id,
            title=task.title,
            messages=[
                TaskMessage.from_event(event)
                for event in session.events
                if event.content
            ],
            state=camelize_dict(session.state),
            extra_info=task.extra_info,
            updated_at=task.updated_at,
        )


class TaskSSEData(BaseModel):
    message: TaskMessage | None = None
    state_delta: dict[str, Any] | None = None
    artifact_delta: dict[str, int] | None = None
    error_code: str | None = None
    error_message: str | None = None

    @classmethod
    def from_event(cls, event: Event) -> "TaskSSEData":
        message = (
            TaskMessage.from_event(event)
            if event.content and event.content.parts
            else None
        )
        state_delta = (
            camelize_dict(event.actions.state_delta) if event.actions else None
        )
        artifact_delta = (
            camelize_dict(event.actions.artifact_delta) if event.actions else None
        )

        return cls(
            message=message,
            state_delta=state_delta,
            artifact_delta=artifact_delta,
            error_code=event.error_code,
            error_message=event.error_message,
        )


class DataSourceContext(BaseModel):
    data_source_id: str
    creator_id: str
    name: str


class InlineContext(BaseModel):
    creator_id: str
    name: str
    content: str


class UserContentWithContext(BaseModel):
    contexts: list[DataSourceContext | InlineContext]
    instruction: str | None = None



from typing import Any, overload

import pydantic
from pydantic.alias_generators import to_camel

base_model_config = {
    "alias_generator": to_camel,
    "validate_by_alias": True,
    "validate_by_name": True,
    "from_attributes": True,
    "arbitrary_types_allowed": True,
}


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(**base_model_config)


class OffsetPagination[ItemT](BaseModel):
    items: list[ItemT]
    total: int
    offset: int
    limit: int


class OrderBy(BaseModel):
    field: str
    desc: bool = True


class ExtraInfoMixin(BaseModel):
    extra_info: dict[str, Any] | None = None

    @overload
    def get_extra_info(self, key: str) -> Any: ...
    @overload
    def get_extra_info[T](self, key: str, default: T) -> T: ...

    def get_extra_info[T](self, key: str, default: T | None = None) -> T | None:
        return self.extra_info.get(key, default) if self.extra_info else default

    @overload
    def pop_extra_info(self, key: str) -> Any: ...
    @overload
    def pop_extra_info[T](self, key: str, default: T) -> T: ...

    def pop_extra_info[T](self, key: str, default: T | None = None) -> T | None:
        return self.extra_info.pop(key, default) if self.extra_info else default

    def set_extra_info(self, key: str, value: Any) -> None:
        if self.extra_info is None:
            self.extra_info = {}
        self.extra_info[key] = value



from uuid import uuid4

from qdrant_client import AsyncQdrantClient
from qdrant_client import models as qdrant_models
from qdrant_client.conversions import common_types as qdrant_types

from sparkden.knowledge.embeddings.base import BaseEmbeddings
from sparkden.shared.utils import getenv

from .base import (
    BaseVectorStore,
    DataChunk,
    DenseOutput,
    Field,
    Filter,
    OrderBy,
    RetrievalMode,
    ScoredDataChunk,
    SparseOutput,
    SparseVector,
    SparseVectorParams,
    VectorParams,
)

_qdrant_client: AsyncQdrantClient | None = None


def get_qdrant_client() -> AsyncQdrantClient:
    global _qdrant_client
    if _qdrant_client is None:
        qdrant_url = getenv("QDRANT_URL")
        _qdrant_client = AsyncQdrantClient(url=qdrant_url)
    return _qdrant_client


class QdrantVectorStore(BaseVectorStore):
    def __init__(self, **kwargs):
        self.client = get_qdrant_client()
        super().__init__(**kwargs)

    @classmethod
    async def create_collection(
        cls,
        collection_name: str,
        *,
        vector_params: VectorParams | None = None,
        sparse_vector_params: SparseVectorParams | None = None,
        payload_fields: list[Field] | None = None,
        recreate_if_exists: bool = False,
    ) -> bool:
        qdrant_client = get_qdrant_client()
        if await qdrant_client.collection_exists(collection_name):
            if recreate_if_exists:
                await qdrant_client.delete_collection(collection_name)
            else:
                return False

        vectors_config = {cls.VECTOR_NAME: vector_params} if vector_params else None

        sparse_vectors_config = (
            {cls.SPARSE_VECTOR_NAME: sparse_vector_params}
            if sparse_vector_params
            else None
        )

        return await qdrant_client.create_collection(
            collection_name=collection_name,
            vectors_config=vectors_config,
            sparse_vectors_config=sparse_vectors_config,
        )

    @classmethod
    async def create_index(
        cls,
        collection_name: str,
        field: Field,
    ) -> None:
        if field.type == "integer":
            field_schema = qdrant_models.IntegerIndexParams(
                type=qdrant_models.IntegerIndexType.INTEGER,
                lookup=False,
                range=True,
            )
        else:
            field_schema = qdrant_models.PayloadSchemaType(field.type)

        qdrant_client = get_qdrant_client()
        await qdrant_client.create_payload_index(
            collection_name=collection_name,
            field_name=field.name,
            field_schema=field_schema,
        )

    async def add_vectors(
        self,
        vectors: DenseOutput | None,
        sparse_vectors: SparseOutput | None,
        payloads: list[dict],
        ids: list[str] | None = None,
    ) -> list[str]:
        if self.retrieval_mode == RetrievalMode.DENSE:
            if vectors is None:
                raise ValueError("Dense vectors are required for dense search mode")
        elif self.retrieval_mode == RetrievalMode.SPARSE:
            if sparse_vectors is None:
                raise ValueError("Sparse vectors are required for sparse search mode")
        elif self.retrieval_mode == RetrievalMode.HYBRID:
            if sparse_vectors is None or vectors is None:
                raise ValueError(
                    "Dense vectors and sparse vectors are required for hybrid search mode"
                )
        else:
            raise ValueError(f"Invalid retrieval mode: {self.retrieval_mode}")

        def get_vector_mapping(idx: int) -> qdrant_models.VectorStruct:
            if self.retrieval_mode == RetrievalMode.DENSE:
                assert vectors is not None
                return {self.VECTOR_NAME: vectors[idx]}
            elif self.retrieval_mode == RetrievalMode.SPARSE:
                assert sparse_vectors is not None
                return {
                    self.SPARSE_VECTOR_NAME: SparseVector(
                        indices=sparse_vectors[idx][0],
                        values=sparse_vectors[idx][1],
                    )
                }
            else:  # HYBRID
                assert vectors is not None and sparse_vectors is not None
                return {
                    self.VECTOR_NAME: vectors[idx],
                    self.SPARSE_VECTOR_NAME: SparseVector(
                        indices=sparse_vectors[idx][0],
                        values=sparse_vectors[idx][1],
                    ),
                }

        if ids is None:
            ids = [str(uuid4().hex) for _ in range(len(payloads))]

        points = [
            qdrant_models.PointStruct(
                id=point_id,
                vector=get_vector_mapping(idx),
                payload=payloads[idx],
            )
            for idx, point_id in enumerate(ids)
        ]
        await self.client.upsert(collection_name=self.collection_name, points=points)

        return ids

    async def count_chunks(self, filter: Filter) -> int:
        return (
            await self.client.count(
                collection_name=self.collection_name,
                count_filter=filter,
            )
        ).count

    async def delete_chunks(
        self, chunk_selector: list[qdrant_types.PointId] | Filter
    ) -> None:
        await self.client.delete(
            collection_name=self.collection_name,
            points_selector=chunk_selector,
        )

    async def filter_chunks(
        self,
        filter: Filter,
        limit: int = 20,
        offset: int = 0,
        order_by: OrderBy | None = None,
    ) -> list[DataChunk]:
        (points, _) = await self.client.scroll(
            collection_name=self.collection_name,
            scroll_filter=filter,
            limit=limit,
            order_by=qdrant_models.OrderBy(
                key=order_by.field,
                direction=qdrant_models.Direction.DESC
                if order_by.desc
                else qdrant_models.Direction.ASC,
                start_from=offset + 1,
            )
            if order_by
            else None,
        )

        return [
            DataChunk(
                id=str(point.id),
                content=point.payload.get("content", "") if point.payload else "",
                metadata={
                    k: v for k, v in (point.payload or {}).items() if k != "content"
                },
            )
            for point in points
        ]

    async def search(
        self,
        query: str,
        embeddings: BaseEmbeddings,
        query_filter: Filter | None = None,
        k: int = 5,
        score_threshold: float | None = None,
    ) -> list[ScoredDataChunk]:
        if self.retrieval_mode == RetrievalMode.DENSE:
            vector = embeddings.embed_query(query, RetrievalMode.DENSE)
            result = await self.client.query_points(
                collection_name=self.collection_name,
                query=vector,
                using=self.VECTOR_NAME,
                query_filter=query_filter,
                limit=k,
                score_threshold=score_threshold,
            )
        elif self.retrieval_mode == RetrievalMode.SPARSE:
            indices, values = embeddings.embed_query(query, RetrievalMode.SPARSE)
            result = await self.client.query_points(
                collection_name=self.collection_name,
                query=qdrant_models.SparseVector(indices=indices, values=values),
                using=self.SPARSE_VECTOR_NAME,
                query_filter=query_filter,
                limit=k,
                score_threshold=score_threshold,
            )
        elif self.retrieval_mode == RetrievalMode.HYBRID:
            vector, sparse_vector = embeddings.embed_query(query, RetrievalMode.HYBRID)
            result = await self.client.query_points(
                collection_name=self.collection_name,
                prefetch=[
                    qdrant_models.Prefetch(
                        query=vector,
                        filter=query_filter,
                        using=self.VECTOR_NAME,
                        limit=round(k * 2 * self.hybrid_dense_ratio),
                        score_threshold=score_threshold,
                    ),
                    qdrant_models.Prefetch(
                        query=qdrant_models.SparseVector(
                            indices=sparse_vector[0],
                            values=sparse_vector[1],
                        ),
                        filter=query_filter,
                        using=self.SPARSE_VECTOR_NAME,
                        limit=round(k * 2 * (1 - self.hybrid_dense_ratio)),
                    ),
                ],
                query=qdrant_models.FusionQuery(fusion=qdrant_models.Fusion.RRF),
                limit=k,
            )
        else:
            raise ValueError(f"Invalid retrieval mode: {self.retrieval_mode}")

        return [
            ScoredDataChunk(
                id=str(point.id),
                content=point.payload.get("content", "") if point.payload else "",
                metadata={
                    k: v for k, v in (point.payload or {}).items() if k != "content"
                },
                score=point.score,
                retrieval_mode=self.retrieval_mode,
            )
            for point in result.points
        ]



from datetime import datetime
from enum import StrEnum
from typing import Any, Awaitable, Callable

from sparkden.models.shared import BaseModel


class QueuedJobStatus(StrEnum):
    PENDING = "pending"
    RUNNING = "running"
    SUCCESS = "success"
    FAILURE = "failure"


JobResult = dict[str, Any] | None

JobFunction = Callable[..., Awaitable[JobResult]]


class QueuedJob(BaseModel):
    id: str
    user_id: str
    status: QueuedJobStatus = QueuedJobStatus.PENDING
    progress: int = 0
    custom_status: str | None = None
    result: JobResult = None
    error: str | None = None
    created_at: datetime
    updated_at: datetime



import json
from abc import ABC
from typing import AsyncGenerator

from qdrant_client.http.models import (
    FieldCondition,
    Filter,
    MatchValue,
)
from sparkden.knowledge.data_parsers.file_parser import FileParser
from sparkden.knowledge.data_splitters.default_splitter import DefaultSplitter
from sparkden.models.knowledge import (
    DataChunk,
    DataSourceCreatorType,
    FileExtractConfig,
    KnowledgeDataSource,
    ParseResult,
    ParseResultType,
    RetrievalMode,
)
from sparkden.shared.minio import (
    delete_objects,
    get_minio_client,
    get_object,
    object_exists,
    save_objects,
)

from .embeddings import Embeddings
from .vector_store import VectorStore


class DataLoader(ABC):
    def __init__(
        self,
        data_source: KnowledgeDataSource,
        vector_dimensions: int = 1024,
        retrieval_mode: RetrievalMode = RetrievalMode.DENSE,
    ):
        from sparkden.knowledge.registry import knowledge_collections

        self.data_source = data_source
        self.vector_dimensions = vector_dimensions
        self.retrieval_mode = retrieval_mode

        data_source_creator = knowledge_collections.get_data_source_creator(
            self.data_source.creator_id
        )
        if not data_source_creator:
            raise ValueError("Data source creator not found")

        self.data_source_creator = data_source_creator

        data_parser = (
            data_source_creator.data_parser
            if data_source_creator.data_parser
            else FileParser()
            if self.data_source_creator.type == DataSourceCreatorType.UPLOAD
            else None
        )

        if not data_parser:
            raise ValueError("Data parser should be provided in data source creator.")

        self.data_parser = data_parser

        self.data_splitter = (
            data_source_creator.data_splitter
            if data_source_creator.data_splitter
            else DefaultSplitter()
        )

    @property
    def bucket_name(self) -> str:
        return self.data_source.collection_id

    @property
    def vector_collection_name(self) -> str:
        return self.data_source.collection_id

    async def parse(self) -> ParseResult:
        object_base_name = self.data_source.id
        parsed_object_name = f"{object_base_name}/parsed"

        # If the data source is already parsed, return the parsed result
        if object_exists(self.bucket_name, parsed_object_name):
            parsed_file = get_object(self.bucket_name, parsed_object_name)
            if not parsed_file:
                raise ValueError(
                    f"Parsed file not found: bucket={self.bucket_name}, object_name={parsed_object_name}"
                )

            if parsed_file.content_type == "text/plain":
                return ParseResult(
                    data=parsed_file.content.decode("utf-8"),
                    data_type=ParseResultType.TEXT,
                )
            elif parsed_file.content_type == "text/markdown":
                return ParseResult(
                    data=parsed_file.content.decode("utf-8"),
                    data_type=ParseResultType.MARKDOWN,
                )
            elif parsed_file.content_type == "application/json":
                return ParseResult(
                    data=json.loads(parsed_file.content.decode("utf-8")),
                    data_type=ParseResultType.TABLE,
                )
            else:
                raise ValueError(
                    f"Unsupported parsed file type: {parsed_file.content_type}"
                )

        if isinstance(self.data_parser, FileParser):
            file = get_object(self.bucket_name, self.data_source.file_object_name)
            if not file:
                raise ValueError(
                    f"File not found: bucket={self.bucket_name}, object_name={self.data_source.file_object_name}"
                )
            input_data = {"file": file}
        else:
            input_data = self.data_source.extra_info

            if not input_data:
                raise ValueError("Data source extra info is required")

        parse_result = await self.data_parser.parse(
            input_data,
            FileExtractConfig(
                object_base_name=object_base_name,
                parsed_object_name=parsed_object_name,
                bucket_name=self.bucket_name,
            ),
        )

        save_objects(self.bucket_name, parse_result.extracted_files)
        return parse_result

    async def load(self, reload: bool = False) -> AsyncGenerator[str, None]:
        if reload:
            yield "unloading"
            await self.unload()

        yield "parsing"
        parse_result = await self.parse()

        yield "splitting"
        chunks = await self.data_splitter.split(parse_result)

        for index, chunk in enumerate(chunks):
            chunk.metadata.update(
                {
                    "sequence_in_data_source": index + 1,
                    "data_source_id": self.data_source.id,
                }
            )

        yield "adding_chunks"
        await self._add_to_vector_store(chunks)

    async def unload(self) -> None:
        minio_client = get_minio_client()
        objects_to_delete = minio_client.list_objects(
            self.data_source.collection_id, prefix=self.data_source.id, recursive=True
        )

        delete_objects(
            self.bucket_name,
            [
                object.object_name
                for object in objects_to_delete
                if object.object_name
                and object.object_name != self.data_source.file_object_name
            ],
        )

        await self._delete_from_vector_store()

    async def reload(self) -> AsyncGenerator[str, None]:
        async for progress in self.load(reload=True):
            yield progress

    async def _add_to_vector_store(self, chunks: list[DataChunk]) -> None:
        embeddings = Embeddings(dimensions=self.vector_dimensions)
        vector_store = VectorStore(
            collection_name=self.vector_collection_name,
            retrieval_mode=self.retrieval_mode,
        )

        await vector_store.add_chunks(chunks, embeddings=embeddings)

    async def _delete_from_vector_store(self) -> None:
        vector_store = VectorStore(
            collection_name=self.vector_collection_name,
        )

        filter = Filter(
            must=[
                FieldCondition(
                    key="data_source_id",
                    match=MatchValue(value=self.data_source.id),
                ),
            ]
        )

        await vector_store.delete_chunks(filter)



import functools

import dramatiq
from dramatiq.middleware import CurrentMessage

from .models import JobFunction, JobResult, QueuedJob
from .service import JobStatusService


class JobContext:
    _service: JobStatusService
    job: QueuedJob

    def __init__(self, service: JobStatusService, job: QueuedJob):
        self._service = service
        self.job = job

    @property
    def job_id(self) -> str:
        return self.job.id

    @property
    def user_id(self) -> str:
        return self.job.user_id

    async def update_progress(
        self,
        progress: int,
        message: str | None = None,
        custom_status: str | None = None,
    ):
        self.job = await self._service.update_job_progress(
            self.job_id, progress, message, custom_status
        )


def queued_job(
    *,
    queue_name: str = "default",
    max_retries: int = 3,
    **dramatiq_options,
):
    def decorator(func: JobFunction) -> JobFunction:
        @dramatiq.actor(
            queue_name=queue_name,
            max_retries=max_retries,
            store_results=True,
            **dramatiq_options,
        )
        @functools.wraps(func)
        async def wrapper(user_id: str | None = None, **kwargs) -> JobResult:
            from sparkden.queued_job.service import get_job_status_service

            service = await get_job_status_service()

            message = CurrentMessage.get_current_message()
            if message is None:
                raise ValueError("Could not get current message")
            job_id = message.message_id
            if user_id is None:
                user_id = "anonymous"

            job = await service.create_job(job_id, user_id)
            context = JobContext(service, job)

            try:
                await service.mark_job_as_running(job_id)

                # Remove user_id from kwargs before passing to the original function
                kwargs.pop("user_id", None)
                # Call the original function with the context
                result = await func(context, **kwargs)

                await service.mark_job_as_success(job_id, result or {})
                return result

            except Exception as e:
                await service.mark_job_as_failure(job_id, str(e))
                raise

        return wrapper

    return decorator



from typing import Literal

from dashscope import TextEmbedding

from .base import (
    BaseEmbeddings,
    DenseOutput,
    HybridOutput,
    RetrievalMode,
    SparseOutput,
)


class QwenEmbeddings(BaseEmbeddings):
    def __init__(self, *, model: str = "text-embedding-v4", dimensions: int = 1024):
        super().__init__(model=model, dimensions=dimensions)

    def _embed_texts(
        self,
        texts: list[str],
        text_type: Literal["document", "query"],
        retrieval_mode: RetrievalMode = RetrievalMode.DENSE,
    ) -> DenseOutput | SparseOutput | HybridOutput:
        response = TextEmbedding.call(
            model=self.model,
            input=texts,
            text_type=text_type,
            output_type="dense&sparse"
            if retrieval_mode == RetrievalMode.HYBRID
            else retrieval_mode,
        )
        if response.status_code != 200:
            raise Exception(
                f"Failed to embed documents: {response.code} - {response.message}"
            )

        embeddings = response.output["embeddings"]
        dense_output: DenseOutput = [
            embedding["embedding"]
            for embedding in embeddings
            if embedding.get("embedding")
        ]

        sparse_output: SparseOutput = [
            (
                [item["index"] for item in embedding["sparse_embedding"]],
                [item["value"] for item in embedding["sparse_embedding"]],
            )
            for embedding in embeddings
            if embedding.get("sparse_embedding")
        ]

        if retrieval_mode == RetrievalMode.DENSE:
            return dense_output
        elif retrieval_mode == RetrievalMode.SPARSE:
            return sparse_output
        elif retrieval_mode == RetrievalMode.HYBRID:
            return (dense_output, sparse_output)



from sparkden.knowledge.embeddings import Embeddings
from sparkden.knowledge.registry import knowledge_collections
from sparkden.knowledge.vector_store import (
    Filter,
    VectorStore,
)
from sparkden.models.knowledge import (
    KnowledgeCollection,
    RetrievalMode,
    ScoredDataChunk,
)
from sparkden.services.base import BaseService


class KnowledgeCollectionService(BaseService):
    async def list_collections(self) -> list[KnowledgeCollection]:
        return knowledge_collections.get_collections()

    async def get_collection(self, collection_id: str) -> KnowledgeCollection | None:
        return knowledge_collections.get_collection(collection_id)

    async def search_collection(
        self,
        collection_id: str,
        query: str,
        filter: Filter | None = None,
        k: int = 5,
        retrieval_mode: RetrievalMode = RetrievalMode.HYBRID,
        score_threshold: float | None = None,
    ) -> list[ScoredDataChunk]:
        collection = knowledge_collections.get_collection(collection_id)
        if not collection:
            raise ValueError(f"Collection not found: {collection_id}")

        vector_store = VectorStore(
            collection_name=collection_id, retrieval_mode=retrieval_mode
        )

        embeddings = Embeddings(dimensions=collection.vector_dimensions)
        return await vector_store.search(query, embeddings, filter, k, score_threshold)



import os
from typing import Any

from litestar.dto import DTOConfig
from pydantic.alias_generators import to_camel, to_snake


def getenv(key: str, default: str | None = None) -> str:
    value = os.getenv(key, default)
    if value is None:
        raise ValueError(f"Environment variable {key} is not defined")
    return value


def camelize_dict(data: dict[str, Any], recursive: bool = True) -> dict[str, Any]:
    if not recursive:
        return {to_camel(key): value for key, value in data.items()}

    def _camelize_value(value: Any) -> Any:
        if isinstance(value, dict):
            return camelize_dict(value, recursive=True)
        elif isinstance(value, list):
            return [_camelize_value(item) for item in value]
        return value

    return {to_camel(key): _camelize_value(value) for key, value in data.items()}


def snake_case_dict(data: dict[str, Any], recursive: bool = True) -> dict[str, Any]:
    if not recursive:
        return {to_snake(key): value for key, value in data.items()}

    def _snake_case_value(value: Any) -> Any:
        if isinstance(value, dict):
            return snake_case_dict(value, recursive=True)
        elif isinstance(value, list):
            return [_snake_case_value(item) for item in value]
        return value

    return {to_snake(key): _snake_case_value(value) for key, value in data.items()}


def dto_config(**kwargs) -> DTOConfig:
    default_kwargs = {
        "rename_strategy": "camel",
        "max_nested_depth": 100,
    }
    return DTOConfig(**default_kwargs, **kwargs)



from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from sparkden.models.knowledge import DataSourceCreator, KnowledgeCollection


class KnowledgeCollectionRegistry:
    """Knowledge collection registry for managing registered knowledge collections."""

    def __init__(self):
        """Initialize the knowledge collection registry."""
        self._collections: dict[str, KnowledgeCollection] = {}

    @overload
    def register(
        self,
        collection: "KnowledgeCollection",
    ) -> "KnowledgeCollection": ...
    @overload
    def register(
        self,
        collection: "list[KnowledgeCollection]",
    ) -> "list[KnowledgeCollection]": ...

    def register(
        self,
        collection: "KnowledgeCollection | list[KnowledgeCollection]",
    ) -> "KnowledgeCollection | list[KnowledgeCollection]":
        """
        Register a new knowledge collection.

        Args:
            collection: The knowledge collection to register

        Returns:
            The registered KnowledgeCollection instance

        Raises:
            ValueError: If a collection with the same ID is already registered
        """
        from sparkden.models.knowledge import DataSourceCreator

        if isinstance(collection, list):
            for c in collection:
                self.register(c)
            return collection

        if collection.id in self._collections:
            raise ValueError(
                f"Knowledge collection with ID '{collection.id}' is already registered"
            )

        if collection.data_source_creators is None:
            collection.data_source_creators = [
                DataSourceCreator(
                    id="file_upload",
                    extra_info={"accept": "application/pdf,text/markdown,text/plain"},
                )
            ]

        self._collections[collection.id] = collection
        return collection

    def get_collection(self, id: str) -> "KnowledgeCollection | None":
        """
        Get a knowledge collection by ID.

        Args:
            id: The collection ID

        Returns:
            The collection if found, None otherwise
        """
        return self._collections.get(id)

    def get_collections(self) -> "list[KnowledgeCollection]":
        """
        Get all registered knowledge collections.

        Returns:
            List of collections sorted by ID for consistent ordering
        """
        return sorted(self._collections.values(), key=lambda c: c.id)

    def unregister(self, id: str) -> bool:
        """
        Unregister a knowledge collection by ID.

        Args:
            id: The collection ID

        Returns:
            True if the collection was found and removed, False otherwise
        """
        if id in self._collections:
            del self._collections[id]
            return True
        return False

    def is_registered(self, id: str) -> bool:
        """
        Check if a knowledge collection is registered.

        Args:
            id: The collection ID

        Returns:
            True if the collection is registered, False otherwise
        """
        return id in self._collections

    def clear(self) -> None:
        """Clear all registered knowledge collections."""
        self._collections.clear()

    def get_data_source_creator(self, creator_id: str) -> "DataSourceCreator | None":
        """
        Get a data source creator by ID from all registered collections.

        Args:
            creator_id: The creator ID

        Returns:
            The data source creator if found, None otherwise
        """
        for collection in self.get_collections():
            if collection.data_source_creators:
                for creator in collection.data_source_creators:
                    if creator.id == creator_id:
                        return creator
        return None

    async def create_all(self) -> None:
        for collection in self.get_collections():
            await create_collection(collection)


async def create_collection(
    knowledge_collection: "KnowledgeCollection",
) -> None:
    from sparkden.knowledge.vector_store import (
        Distance,
        Field,
        SparseVectorParams,
        VectorParams,
        VectorStore,
    )
    from sparkden.shared.minio import get_minio_client

    try:
        minio_client = get_minio_client()
        if not minio_client.bucket_exists(knowledge_collection.id):
            minio_client.make_bucket(knowledge_collection.id)
            print(f"Created MinIO bucket: {knowledge_collection.id}")
    except Exception as e:
        print(f"Unexpected error initializing MinIO bucket: {e}")
        raise

    try:
        created = await VectorStore.create_collection(
            knowledge_collection.id,
            vector_params=VectorParams(
                size=knowledge_collection.vector_dimensions,
                distance=Distance(knowledge_collection.vector_distance),
            ),
            sparse_vector_params=SparseVectorParams(),
        )
        if created:
            print(f"Vector store collection created: {knowledge_collection.id}")

            await VectorStore.create_index(
                knowledge_collection.id,
                Field(name="sequence_in_data_source", type="integer"),
            )
    except Exception as e:
        print(f"Unexpected error initializing vector store collection: {e}")
        raise


knowledge_collections = KnowledgeCollectionRegistry()
knowledge_collection_registry = knowledge_collections



import hashlib
from typing import Annotated, Any

from bcrypt import checkpw
from litestar import Controller, Request, delete, get, post
from litestar.connection import ASGIConnection
from litestar.exceptions import NotAuthorizedException
from litestar.middleware.authentication import AuthenticationResult
from litestar.middleware.session.server_side import (
    ServerSideSessionBackend,
    ServerSideSessionConfig,
)
from litestar.plugins.pydantic import PydanticDTO
from litestar.security.session_auth import SessionAuth, SessionAuthMiddleware
from sparkden.db.schema import ApiKey
from sparkden.db.schema import User as StorageUser
from sparkden.models.shared import BaseModel
from sparkden.models.user import User
from sparkden.shared.pg import get_session_maker
from sparkden.shared.utils import dto_config
from sqlalchemy import select


class Session(BaseModel):
    user: User


class SignInUser(BaseModel):
    username: str
    password: str


SignInUserDTO = PydanticDTO[Annotated[SignInUser, dto_config()]]
SessionDTO = PydanticDTO[Annotated[Session, dto_config()]]


class SessionController(Controller):
    path = "/sessions"

    @get("/", return_dto=SessionDTO)
    async def get_session(self, request: Request) -> Session:
        return Session(user=request.user)

    @post("/", dto=SignInUserDTO, return_dto=SessionDTO, exclude_from_auth=True)
    async def sign_in(self, data: SignInUser, request: Request) -> Session:
        async with get_session_maker().begin() as db_session:
            user = await db_session.scalar(
                select(StorageUser).where(StorageUser.username == data.username)
            )
            if not user or not user.password:
                raise NotAuthorizedException(
                    status_code=401, detail="Invalid username or password"
                )
            if not checkpw(data.password.encode(), user.password):
                raise NotAuthorizedException(
                    status_code=401, detail="Invalid username or password"
                )
            request.set_session({"user_id": user.id})
            return Session(user=User.from_storage(user))

    @delete("/", exclude_from_auth=True)
    async def sign_out(self, request: Request) -> None:
        request.clear_session()


async def retrieve_user_handler(
    session: dict[str, Any], connection: ASGIConnection
) -> User | None:
    async with get_session_maker().begin() as db_session:
        if user_id := session.get("user_id"):
            storage_user = await db_session.get(StorageUser, user_id)
        return User.from_storage(storage_user) if storage_user else None


class APIKeyAndSessionAuthMiddleware(SessionAuthMiddleware):
    async def authenticate_request(
        self, connection: ASGIConnection[Any, Any, Any, Any]
    ) -> AuthenticationResult:
        if api_key := connection.headers.get("X-API-KEY"):
            key_hash = hashlib.sha256(api_key.encode()).hexdigest()
            async with get_session_maker().begin() as db_session:
                storage_user = await db_session.scalar(
                    select(StorageUser).join(ApiKey).where(ApiKey.key_hash == key_hash)
                )
                if storage_user:
                    return AuthenticationResult(
                        user=User.from_storage(storage_user), auth=connection.session
                    )
        return await super().authenticate_request(connection)


session_backend_config = ServerSideSessionConfig()

session_auth = SessionAuth[User, ServerSideSessionBackend](
    retrieve_user_handler=retrieve_user_handler,
    session_backend_config=session_backend_config,
    authentication_middleware_class=APIKeyAndSessionAuthMiddleware,
    route_handlers=[
        SessionController,
    ],
)



from abc import ABC
from enum import StrEnum
from typing import Callable, NotRequired, Sequence, TypedDict

from google.adk.agents import BaseAgent
from google.adk.plugins.base_plugin import BasePlugin
from litestar import Controller
from litestar.datastructures import UploadFile
from litestar.handlers import ASGIRouteHandler
from pydantic import Field, field_serializer
from sparkden.assistants.context_selector import ContextSelector
from sparkden.models.knowledge import DataSourceCreator
from sparkden.models.shared import BaseModel, ExtraInfoMixin


class ToolResponseStatus(StrEnum):
    PENDING = "pending"
    SUCCESS = "success"
    ERROR = "error"


class ToolResponse[ResultT](TypedDict):
    """The base tool response."""

    result: NotRequired[ResultT]
    """The result of the tool."""

    status: NotRequired[ToolResponseStatus]
    """The status of the tool response."""

    error: NotRequired[str]
    """The error message of the tool response."""


class UserApprovalResult[ItemT](BaseModel):
    """The user approval result."""

    approved: bool
    """Whether the user approved the request."""

    feedback: str
    """The feedback for the user approval."""

    modified_item: ItemT | None = None
    """A modified item provided on approval as result, or on rejection as a suggestion."""


class ProgressItemStatus(StrEnum):
    COMPLETED = "completed"
    RUNNING = "running"
    PENDING = "pending"


AssistantApiRoute = type[Controller] | ASGIRouteHandler


class ContextOption(ExtraInfoMixin, BaseModel):
    id: str
    name: str
    description: str | None = None


class ContextWithFile(ContextOption):
    creator_id: str
    file: UploadFile | None = None


class ContextCreator(DataSourceCreator, ABC):
    context_selector: ContextSelector | None = None

    @field_serializer("context_selector")
    def serialize_context_selector(
        self, context_selector: ContextSelector | None
    ) -> bool:
        return context_selector is not None


class Assistant(BaseModel):
    id: str
    disabled: bool = False
    root_agent: BaseAgent = Field(exclude=True)
    api_routes: Sequence[AssistantApiRoute] | None = Field(default=None, exclude=True)
    agent_plugins: list[BasePlugin] = Field(default_factory=list, exclude=True)
    callbacks: dict[str, Callable] | None = Field(default=None, exclude=True)
    context_creators: list[ContextCreator] | None = None
    search_knowledge: str | None = None
    sequence: int = 9999

    def get_context_creator(self, creator_id: str) -> ContextCreator | None:
        return next(
            (
                context_creator
                for context_creator in self.context_creators or []
                if context_creator.id == creator_id
            ),
            None,
        )



import os
from typing import TYPE_CHECKING

from dotenv import load_dotenv
from litestar import Litestar

if TYPE_CHECKING:
    from .models.assistant import Assistant
    from .models.knowledge import KnowledgeCollection

# env variables auto loaded in docker compose
if os.getenv("REDIS_URL") is None:
    load_dotenv()


def create_app(
    assistants: "list[Assistant]" = [],
    knowledge_collections: "list[KnowledgeCollection]" = [],
) -> Litestar:
    import dramatiq
    from litestar.config.compression import CompressionConfig
    from litestar.config.cors import CORSConfig
    from litestar.config.csrf import CSRFConfig
    from litestar.stores.redis import RedisStore

    from .assistants.registry import assistant_registry
    from .db.init import init_db
    from .knowledge.registry import knowledge_collection_registry
    from .routes.assistants import AssistantController
    from .routes.files import FilesController
    from .routes.knowledge import KnowledgeController
    from .routes.queued_jobs import QueuedJobController
    from .routes.security.auth import session_auth, session_backend_config
    from .routes.security.csrf import csrf_token
    from .routes.tasks import TaskController
    from .shared import getenv
    from .shared.pg import dispose_engine
    from .shared.redis import get_async_redis_client
    from .worker.broker import create_broker

    broker = create_broker()
    dramatiq.set_broker(broker)

    assistant_registry.register(assistants)
    knowledge_collection_registry.register(knowledge_collections)

    return Litestar(
        route_handlers=[
            csrf_token,
            AssistantController,
            TaskController,
            KnowledgeController,
            FilesController,
            QueuedJobController,
            *assistant_registry.get_api_routes(),
        ],
        stores={
            session_backend_config.store: RedisStore(
                redis=get_async_redis_client(),
                namespace="sessions",
                handle_client_shutdown=True,
            ),
        },
        lifespan=[*assistant_registry.get_callbacks("lifespan")],
        on_app_init=[
            session_auth.on_app_init,
            *assistant_registry.get_callbacks("on_app_init"),
        ],
        on_startup=[
            init_db,
            knowledge_collection_registry.create_all,
            *assistant_registry.get_callbacks("on_startup"),
        ],
        on_shutdown=[dispose_engine, *assistant_registry.get_callbacks("on_shutdown")],
        csrf_config=CSRFConfig(
            secret=getenv("CSRF_SECRET"),
        ),
        cors_config=CORSConfig(
            allow_origins=[url.strip() for url in getenv("CLIENT_URL", "*").split(",")],
            allow_credentials=True,
        ),
        compression_config=CompressionConfig(backend="gzip"),
        openapi_config=None,
        debug=getenv("DEBUG", "false") == "true",
    )



from typing import TYPE_CHECKING, Callable, Sequence, overload

if TYPE_CHECKING:
    from sparkden.models.assistant import Assistant, AssistantApiRoute


class AssistantRegistry:
    """Assistant registry for managing registered assistants."""

    def __init__(self):
        """Initialize the assistant registry."""
        self._assistants: dict[str, Assistant] = {}

    @overload
    def register(
        self,
        assistant: "Assistant",
    ) -> "Assistant": ...
    @overload
    def register(
        self,
        assistant: "list[Assistant]",
    ) -> "list[Assistant]": ...

    def register(
        self,
        assistant: "Assistant | list[Assistant]",
    ) -> "Assistant | list[Assistant]":
        """
        Register a new assistant.

        Args:
            assistant: The assistant to register

        Returns:
            The registered Assistant instance

        Raises:
            ValueError: If an assistant with the same ID is already registered
        """
        from google.adk.agents import LlmAgent

        from sparkden.assistants.helpers import assistant_context_collection_id
        from sparkden.assistants.tools.search_context_tool import search_context
        from sparkden.assistants.tools.search_knowledge_tool import SearchKnowledgeTool
        from sparkden.knowledge.registry import knowledge_collections
        from sparkden.models.knowledge import KnowledgeCollection

        if isinstance(assistant, list):
            for a in assistant:
                self.register(a)
            return assistant

        if assistant.id in self._assistants:
            raise ValueError(
                f"Assistant with ID '{assistant.id}' is already registered"
            )

        self._assistants[assistant.id] = assistant

        if assistant.context_creators:
            knowledge_collections.register(
                KnowledgeCollection(
                    id=assistant_context_collection_id(assistant.id),
                    data_source_creators=assistant.context_creators,
                )
            )

            if isinstance(assistant.root_agent, LlmAgent):
                assistant.root_agent.tools.append(search_context)

        if assistant.search_knowledge and isinstance(assistant.root_agent, LlmAgent):
            assistant.root_agent.tools.append(
                SearchKnowledgeTool(collection_id=assistant.search_knowledge)
            )

        return assistant

    def get_assistant(self, id: str) -> "Assistant | None":
        """
        Get an assistant by ID.

        Args:
            id: The assistant ID

        Returns:
            The assistant if found, None otherwise
        """
        return self._assistants.get(id)

    def get_assistants(self, include_disabled: bool = False) -> "list[Assistant]":
        """
        Get all registered assistants.

        Args:
            include_disabled: Whether to include disabled assistants (default: True)

        Returns:
            List of assistants sorted by sequence
        """
        assistants = []
        if include_disabled:
            assistants = list(self._assistants.values())
        else:
            assistants = [a for a in self._assistants.values() if not a.disabled]

        # Sort by sequence, then by ID for consistent ordering
        return sorted(assistants, key=lambda a: (a.sequence, a.id))

    def unregister(self, id: str) -> bool:
        """
        Unregister an assistant by ID.

        Args:
            id: The assistant ID

        Returns:
            True if the assistant was found and removed, False otherwise
        """
        if id in self._assistants:
            del self._assistants[id]
            return True
        return False

    def is_registered(self, id: str) -> bool:
        """
        Check if an assistant is registered.

        Args:
            id: The assistant ID

        Returns:
            True if the assistant is registered, False otherwise
        """
        return id in self._assistants

    def enable_assistant(self, id: str) -> bool:
        """
        Enable an assistant by ID.

        Args:
            id: The assistant ID

        Returns:
            True if the assistant was found and enabled, False otherwise
        """
        assistant = self._assistants.get(id)
        if assistant:
            assistant.disabled = False
            return True
        return False

    def disable_assistant(self, id: str) -> bool:
        """
        Disable an assistant by ID.

        Args:
            id: The assistant ID

        Returns:
            True if the assistant was found and disabled, False otherwise
        """
        assistant = self._assistants.get(id)
        if assistant:
            assistant.disabled = True
            return True
        return False

    def update_sequence(self, id: str, sequence: int) -> bool:
        """
        Update the sequence of an assistant.

        Args:
            id: The assistant ID
            sequence: The new sequence value

        Returns:
            True if the assistant was found and updated, False otherwise
        """
        assistant = self._assistants.get(id)
        if assistant:
            assistant.sequence = sequence
            return True
        return False

    def clear(self) -> None:
        """Clear all registered assistants."""
        self._assistants.clear()

    def get_api_routes(self) -> "Sequence[AssistantApiRoute]":
        routes: "Sequence[AssistantApiRoute]" = []
        for assistant in self.get_assistants():
            if assistant.api_routes:
                routes.extend(assistant.api_routes)
        return routes

    def get_callbacks(self, callback_name: str) -> list[Callable]:
        callbacks = []
        for assistant in self.get_assistants():
            callback = (assistant.callbacks or {}).get(callback_name)
            if callback is not None:
                callbacks.append(callback)
        return callbacks


assistant_registry = AssistantRegistry()



def create_broker():
    from dramatiq.brokers.redis import RedisBroker
    from dramatiq.middleware import AsyncIO, CurrentMessage
    from dramatiq.results import Results
    from dramatiq.results.backends.redis import RedisBackend

    from sparkden.shared.redis import get_redis_client

    redis_client = get_redis_client()
    result_backend = RedisBackend(client=redis_client)

    # This broker instance will be discovered by the dramatiq CLI

    broker = RedisBroker(client=redis_client)
    broker.add_middleware(AsyncIO())
    broker.add_middleware(Results(backend=result_backend))
    broker.add_middleware(CurrentMessage())
    return broker



import logging
from datetime import datetime
from typing import Any, cast, override

from google.adk.events import Event
from google.adk.sessions import Session
from google.adk.sessions.base_session_service import (
    BaseSessionService,
    GetSessionConfig,
    ListSessionsResponse,
)
from google.adk.sessions.database_session_service import (
    StorageAppState,
    StorageEvent,
    StorageSession,
    StorageUserState,
    _extract_state_delta,
    _merge_state,
)
from sparkden.shared.pg import get_session_maker
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession
from tzlocal import get_localzone

logger = logging.getLogger(__name__)


class SessionService(BaseSessionService):
    """A session service that uses a database for storage."""

    def __init__(self):
        # Get the local timezone
        local_timezone = get_localzone()
        logger.info(f"Local timezone: {local_timezone}")

    @override
    async def create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: dict[str, Any] | None = None,
        session_id: str | None = None,
        db_session: AsyncSession | None = None,
    ) -> Session:
        if db_session:
            return await self._create_session(
                app_name=app_name,
                user_id=user_id,
                state=state,
                session_id=session_id,
                db_session=db_session,
            )
        else:
            async with get_session_maker().begin() as db_session:
                return await self._create_session(
                    app_name=app_name,
                    user_id=user_id,
                    state=state,
                    session_id=session_id,
                    db_session=db_session,
                )

    async def _create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: dict[str, Any] | None = None,
        session_id: str | None = None,
        db_session: AsyncSession,
    ) -> Session:
        # 1. Populate states.
        # 2. Build storage session object
        # 3. Add the object to the table
        # 4. Build the session object with generated id
        # 5. Return the session

        # Fetch app and user states from storage
        storage_app_state = await db_session.get(StorageAppState, (app_name))
        storage_user_state = await db_session.get(StorageUserState, (app_name, user_id))

        app_state = storage_app_state.state if storage_app_state else {}
        user_state = storage_user_state.state if storage_user_state else {}

        # Create state tables if not exist
        if not storage_app_state:
            storage_app_state = StorageAppState(app_name=app_name, state={})
            db_session.add(storage_app_state)
        if not storage_user_state:
            storage_user_state = StorageUserState(
                app_name=app_name, user_id=user_id, state={}
            )
            db_session.add(storage_user_state)

        # Extract state deltas
        app_state_delta, user_state_delta, session_state = _extract_state_delta(
            state  # type: ignore
        )

        # Apply state delta
        app_state.update(app_state_delta)
        user_state.update(user_state_delta)

        # Store app and user state
        if app_state_delta:
            storage_app_state.state = app_state  # type: ignore
        if user_state_delta:
            storage_user_state.state = user_state  # type: ignore

        # Store the session
        storage_session = StorageSession(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
            state=session_state,
        )
        db_session.add(storage_session)
        await db_session.flush()
        await db_session.refresh(storage_session)

        # Merge states for response
        merged_state = _merge_state(app_state, user_state, session_state)
        session = storage_session.to_session(state=merged_state)
        return session

    @override
    async def get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: GetSessionConfig | None = None,
        db_session: AsyncSession | None = None,
    ) -> Session | None:
        if db_session:
            return await self._get_session(
                app_name=app_name,
                user_id=user_id,
                session_id=session_id,
                config=config,
                db_session=db_session,
            )
        else:
            async with get_session_maker().begin() as db_session:
                return await self._get_session(
                    app_name=app_name,
                    user_id=user_id,
                    session_id=session_id,
                    config=config,
                    db_session=db_session,
                )

    async def _get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: GetSessionConfig | None = None,
        db_session: AsyncSession,
    ) -> Session | None:
        # 1. Get the storage session entry from session table
        # 2. Get all the events based on session id and filtering config
        # 3. Convert and return the session
        storage_session = await db_session.get(
            StorageSession, (app_name, user_id, session_id)
        )
        if storage_session is None:
            return None

        conditions = [
            StorageEvent.app_name == app_name,
            StorageEvent.session_id == storage_session.id,
            StorageEvent.user_id == user_id,
        ]

        if config and config.after_timestamp:
            after_dt = datetime.fromtimestamp(config.after_timestamp)
            conditions.append(StorageEvent.timestamp >= after_dt)

        stmt = (
            select(StorageEvent)
            .where(*conditions)
            .order_by(StorageEvent.timestamp.desc())
        )

        if config and config.num_recent_events:
            stmt = stmt.limit(config.num_recent_events)  # type: ignore

        storage_events = (await db_session.scalars(stmt)).all()

        # Fetch states from storage
        storage_app_state = await db_session.get(StorageAppState, app_name)
        storage_user_state = await db_session.get(StorageUserState, (app_name, user_id))

        app_state = storage_app_state.state if storage_app_state else {}
        user_state = storage_user_state.state if storage_user_state else {}
        session_state = storage_session.state

        # Merge states
        merged_state = _merge_state(app_state, user_state, session_state)

        # Convert storage session to session
        events = [e.to_event() for e in reversed(storage_events)]
        session = storage_session.to_session(state=merged_state, events=events)
        return session

    @override
    async def list_sessions(
        self,
        *,
        app_name: str,
        user_id: str,
        db_session: AsyncSession | None = None,
    ) -> ListSessionsResponse:
        if db_session:
            return await self._list_sessions(
                app_name=app_name,
                user_id=user_id,
                db_session=db_session,
            )
        else:
            async with get_session_maker().begin() as db_session:
                return await self._list_sessions(
                    app_name=app_name,
                    user_id=user_id,
                    db_session=db_session,
                )

    async def _list_sessions(
        self,
        *,
        app_name: str,
        user_id: str,
        db_session: AsyncSession,
    ) -> ListSessionsResponse:
        stmt = select(StorageSession).where(
            StorageSession.app_name == app_name,
            StorageSession.user_id == user_id,
        )
        results = (await db_session.scalars(stmt)).all()

        # Fetch states from storage
        storage_app_state = await db_session.get(StorageAppState, app_name)
        storage_user_state = await db_session.get(StorageUserState, (app_name, user_id))

        app_state = storage_app_state.state if storage_app_state else {}
        user_state = storage_user_state.state if storage_user_state else {}

        sessions = []
        for storage_session in results:
            session_state = storage_session.state
            merged_state = _merge_state(app_state, user_state, session_state)
            sessions.append(storage_session.to_session(state=merged_state))

        return ListSessionsResponse(sessions=sessions)

    @override
    async def delete_session(
        self,
        app_name: str,
        user_id: str,
        session_id: str,
        db_session: AsyncSession | None = None,
    ) -> None:
        if db_session:
            return await self._delete_session(
                app_name=app_name,
                user_id=user_id,
                session_id=session_id,
                db_session=db_session,
            )
        else:
            async with get_session_maker().begin() as db_session:
                return await self._delete_session(
                    app_name=app_name,
                    user_id=user_id,
                    session_id=session_id,
                    db_session=db_session,
                )

    async def _delete_session(
        self,
        app_name: str,
        user_id: str,
        session_id: str,
        db_session: AsyncSession,
    ) -> None:
        stmt = delete(StorageSession).where(
            StorageSession.app_name == app_name,
            StorageSession.user_id == user_id,
            StorageSession.id == session_id,
        )
        await db_session.execute(stmt)

    @override
    async def append_event(
        self, session: Session, event: Event, db_session: AsyncSession | None = None
    ) -> Event:
        if db_session:
            return await self._append_event(
                session=session,
                event=event,
                db_session=db_session,
            )
        else:
            async with get_session_maker().begin() as db_session:
                return await self._append_event(
                    session=session,
                    event=event,
                    db_session=db_session,
                )

    async def _append_event(
        self, session: Session, event: Event, db_session: AsyncSession
    ) -> Event:
        logger.info(f"Append event: {event} to session {session.id}")

        if event.partial:
            return event

        # 1. Check if timestamp is stale
        # 2. Update session attributes based on event config
        # 3. Store event to table
        storage_session = await db_session.get(
            StorageSession, (session.app_name, session.user_id, session.id)
        )

        if storage_session is None:
            raise ValueError(f"Session {session.id} not found")

        session_update_time = cast(datetime, storage_session.update_time).timestamp()

        if session_update_time > session.last_update_time:
            raise ValueError(
                "The last_update_time provided in the session object"
                f" {datetime.fromtimestamp(session.last_update_time):'%Y-%m-%d %H:%M:%S'} is"
                " earlier than the update_time in the storage_session"
                f" {datetime.fromtimestamp(session_update_time):'%Y-%m-%d %H:%M:%S'}."
                " Please check if it is a stale session."
            )

        # Fetch states from storage
        storage_app_state = await db_session.get(StorageAppState, session.app_name)
        storage_user_state = await db_session.get(
            StorageUserState, (session.app_name, session.user_id)
        )

        app_state = storage_app_state.state if storage_app_state else {}
        user_state = storage_user_state.state if storage_user_state else {}
        session_state = storage_session.state

        # Extract state delta
        app_state_delta = {}
        user_state_delta = {}
        session_state_delta = {}
        if event.actions and event.actions.state_delta:
            app_state_delta, user_state_delta, session_state_delta = (
                _extract_state_delta(event.actions.state_delta)
            )

        # Merge state and update storage
        if app_state_delta:
            app_state.update(app_state_delta)
            storage_app_state.state = app_state  # type: ignore
        if user_state_delta:
            user_state.update(user_state_delta)
            storage_user_state.state = user_state  # type: ignore
        if session_state_delta:
            session_state.update(session_state_delta)
            storage_session.state = session_state

        db_session.add(StorageEvent.from_event(session, event))
        await db_session.flush()
        await db_session.refresh(storage_session)

        # Update timestamp with commit time
        session.last_update_time = cast(
            datetime, storage_session.update_time
        ).timestamp()

        # Also update the in-memory session
        await super().append_event(session=session, event=event)
        return event



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Type

from sparkden.models.shared import BaseModel

if TYPE_CHECKING:
    from sparkden.models.knowledge import FileExtractConfig, ParseResult


class BaseDataParser[DataType: BaseModel](ABC):
    def __init__(self, input_data_type: Type[DataType]):
        self.input_data_type = input_data_type

    def _parse_input_data(self, input_data: dict[str, Any] | DataType) -> DataType:
        if isinstance(input_data, dict):
            return self.input_data_type.model_validate(input_data)
        else:
            return input_data

    @abstractmethod
    async def parse(
        self,
        input_data: dict[str, Any] | DataType,
        file_extract_config: "FileExtractConfig",
    ) -> "ParseResult":
        pass



import asyncio
from typing import Annotated, AsyncGenerator

from litestar import Controller, get
from litestar.di import Provide
from litestar.plugins.pydantic import PydanticDTO
from litestar.response import ServerSentEvent, ServerSentEventMessage
from sparkden.models.shared import BaseModel
from sparkden.queued_job.models import QueuedJob, QueuedJobStatus
from sparkden.queued_job.service import JobStatusService, get_job_status_service
from sparkden.shared.utils import dto_config


class GetQueuedJobResponse(BaseModel):
    job: QueuedJob | None = None
    error: str | None = None


GetQueuedJobResponseDTO = PydanticDTO[Annotated[GetQueuedJobResponse, dto_config()]]


async def provide_job_status_service() -> JobStatusService:
    return await get_job_status_service()


class QueuedJobController(Controller):
    path = "/jobs"
    dependencies = {
        "job_service": Provide(provide_job_status_service),
    }

    @get("/{job_id:str}", return_dto=GetQueuedJobResponseDTO)
    async def get_job(
        self, job_service: JobStatusService, job_id: str
    ) -> GetQueuedJobResponse:
        job = await job_service.get_job(job_id)
        return GetQueuedJobResponse(job=job)

    @get("/{job_id:str}/progress")
    async def job_events(
        self, job_service: JobStatusService, job_id: str
    ) -> ServerSentEvent:
        async def event_generator() -> AsyncGenerator[ServerSentEventMessage, None]:
            # Poll for the job to appear, with a timeout.
            initial_job = None
            for _ in range(10):  # Poll for 10 seconds
                initial_job = await job_service.get_job(job_id)
                if initial_job:
                    break
                await asyncio.sleep(1)

            if not initial_job:
                yield ServerSentEventMessage(
                    data=GetQueuedJobResponse(
                        error="Job not found",
                    ).model_dump_json(exclude_none=True, by_alias=True),
                    event="message",
                )
                return

            yield ServerSentEventMessage(
                data=GetQueuedJobResponse(
                    job=initial_job,
                ).model_dump_json(exclude_none=True, by_alias=True),
                event="message",
            )

            if initial_job.status in [QueuedJobStatus.SUCCESS, QueuedJobStatus.FAILURE]:
                return

            # Then, subscribe to updates
            pubsub = job_service._redis.pubsub()
            await pubsub.subscribe(job_service._get_pubsub_channel(job_id))

            while True:
                try:
                    message = await pubsub.get_message(
                        ignore_subscribe_messages=True, timeout=10
                    )
                    if message and message["type"] == "message":
                        data = message["data"]
                        job_data = QueuedJob.model_validate_json(data)
                        yield ServerSentEventMessage(
                            data=GetQueuedJobResponse(
                                job=job_data,
                            ).model_dump_json(exclude_none=True, by_alias=True),
                            event="message",
                        )
                        # If the job is finished, break the loop
                        if job_data.status in [
                            QueuedJobStatus.SUCCESS,
                            QueuedJobStatus.FAILURE,
                        ]:
                            break
                    await asyncio.sleep(1)
                except asyncio.CancelledError:
                    await pubsub.unsubscribe(job_service._get_pubsub_channel(job_id))
                    break

        return ServerSentEvent(content=event_generator())



import json
from typing import Annotated

from litestar import Controller, Request, delete, get, post
from litestar.connection import ASGIConnection
from litestar.datastructures import UploadFile
from litestar.di import Provide
from litestar.enums import RequestEncodingType
from litestar.exceptions import (
    NotAuthorizedException,
    NotFoundException,
    ValidationException,
)
from litestar.handlers.base import BaseRouteHandler
from litestar.params import Body
from litestar.plugins.pydantic import PydanticDTO
from sparkden.knowledge.registry import knowledge_collections
from sparkden.knowledge.vector_store.base import FieldCondition, Filter, MatchAny
from sparkden.models.shared import BaseModel, OffsetPagination
from sparkden.services.knowledge import KnowledgeService
from sparkden.shared.utils import dto_config

from ..models.knowledge import (
    DataChunk,
    DataSourceCreatorType,
    FileObject,
    KnowledgeCollection,
    KnowledgeDataSource,
    KnowledgeDataSourceCreate,
    RetrievalMode,
    ScoredDataChunk,
)


class CollectionsResponse(BaseModel):
    collections: list[KnowledgeCollection]


class CreateDataSourceParams(BaseModel):
    creator_id: str
    extra_info: str | None = None
    file: UploadFile | None = None


class CreateDataSourceResponse(BaseModel):
    data_source: KnowledgeDataSource
    created: bool
    job_id: str | None = None


class RefreshDataSourceResponse(BaseModel):
    job_id: str


class SearchCollectionResponse(BaseModel):
    chunks: list[ScoredDataChunk]


CollectionsResponseDTO = PydanticDTO[
    Annotated[
        CollectionsResponse,
        dto_config(),
    ]
]

CollectionResponseDTO = PydanticDTO[
    Annotated[
        KnowledgeCollection,
        dto_config(),
    ]
]

ListDataSourcesResponseDTO = PydanticDTO[
    Annotated[
        OffsetPagination[KnowledgeDataSource],
        dto_config(),
    ]
]

CreateDataSourceResponseDTO = PydanticDTO[
    Annotated[
        CreateDataSourceResponse,
        dto_config(),
    ]
]

RefreshDataSourceResponseDTO = PydanticDTO[
    Annotated[
        RefreshDataSourceResponse,
        dto_config(),
    ]
]

CreateDataSourceParamsDTO = PydanticDTO[
    Annotated[
        CreateDataSourceParams,
        dto_config(),
    ]
]

ListDataChunksResponseDTO = PydanticDTO[
    Annotated[
        OffsetPagination[DataChunk],
        dto_config(),
    ]
]

SearchResultsResponseDTO = PydanticDTO[
    Annotated[
        SearchCollectionResponse,
        dto_config(),
    ]
]


def get_knowledge_service(request: Request) -> KnowledgeService:
    return KnowledgeService(
        user_id=str(request.user.id),
    )


def admin_user_guard(connection: ASGIConnection, _: BaseRouteHandler) -> None:
    if not connection.user.is_admin:
        raise NotAuthorizedException()


class KnowledgeController(Controller):
    path = "/knowledge"
    guards = [admin_user_guard]
    dependencies = {
        "knowledge_service": Provide(get_knowledge_service, sync_to_thread=True),
    }
    request_max_body_size = 50 * 1024 * 1024  # 50MB

    @get(
        "/collections",
        return_dto=CollectionsResponseDTO,
    )
    async def list_collections(
        self, knowledge_service: KnowledgeService
    ) -> CollectionsResponse:
        collections = await knowledge_service.list_collections()
        return CollectionsResponse(collections=collections)

    @get(
        "/collections/{collection_id:str}",
        return_dto=CollectionResponseDTO,
    )
    async def get_collection(
        self, collection_id: str, knowledge_service: KnowledgeService
    ) -> KnowledgeCollection:
        collection = await knowledge_service.get_collection(collection_id=collection_id)
        if not collection:
            raise NotFoundException(status_code=404, detail="Collection not found")
        return collection

    @get(
        "/collections/{collection_id:str}/data-sources",
        return_dto=ListDataSourcesResponseDTO,
    )
    async def list_data_sources(
        self,
        collection_id: str,
        knowledge_service: KnowledgeService,
        limit: int = 30,
        offset: int = 0,
    ) -> OffsetPagination[KnowledgeDataSource]:
        return await knowledge_service.list_data_sources(
            collection_id=collection_id, limit=limit, offset=offset
        )

    @get(
        "/collections/{collection_id:str}/data-sources/{data_source_id:str}/chunks",
        return_dto=ListDataChunksResponseDTO,
    )
    async def list_data_source_chunks(
        self,
        collection_id: str,
        data_source_id: str,
        knowledge_service: KnowledgeService,
        limit: int = 20,
        offset: int = 0,
    ) -> OffsetPagination[DataChunk]:
        return await knowledge_service.list_data_source_chunks(
            collection_id,
            data_source_id,
            limit=limit,
            offset=offset,
        )

    @post(
        "/collections/{collection_id:str}/data-sources",
        return_dto=CreateDataSourceResponseDTO,
        dto=CreateDataSourceParamsDTO,
    )
    async def create_data_source(
        self,
        collection_id: str,
        knowledge_service: KnowledgeService,
        data: Annotated[
            CreateDataSourceParams, Body(media_type=RequestEncodingType.MULTI_PART)
        ],
    ) -> CreateDataSourceResponse:
        data_source_creator = knowledge_collections.get_data_source_creator(
            data.creator_id
        )
        if not data_source_creator:
            raise ValidationException(
                status_code=400, detail="Data source creator not found"
            )

        if data_source_creator.type == DataSourceCreatorType.UPLOAD:
            if not data.file:
                raise ValidationException(status_code=400, detail="File is required")
            accept = [
                mime.strip()
                for mime in data_source_creator.get_extra_info("accept", "").split(",")
            ]
            if not any(data.file.content_type.startswith(mime) for mime in accept):
                raise ValidationException(
                    status_code=400,
                    detail=f"File must be one of the following types: {', '.join(accept)}",
                )
            file_content = await data.file.read()
            data_source = KnowledgeDataSourceCreate(
                creator_id=data_source_creator.id,
                name=data.file.filename,
                collection_id=collection_id,
                extra_info={
                    **(data_source_creator.extra_info or {}),
                    **(json.loads(data.extra_info) if data.extra_info else {}),
                    "file": FileObject(
                        name=data.file.filename,
                        content=file_content,
                        content_type=data.file.content_type,
                    ),
                },
            )
            data_source, created, job_id = await knowledge_service.add_data_source(
                data_source=data_source
            )
            return CreateDataSourceResponse(
                data_source=data_source, created=created, job_id=job_id
            )
        else:
            # TODO: create remote data source
            raise NotImplementedError("Not implemented")

    @delete("/collections/{collection_id:str}/data-sources/{data_source_id:str}")
    async def delete_data_source(
        self, data_source_id: str, knowledge_service: KnowledgeService
    ) -> None:
        await knowledge_service.delete_data_source(data_source_id)

    @post(
        "/collections/{collection_id:str}/data-sources/{data_source_id:str}/refresh",
        return_dto=RefreshDataSourceResponseDTO,
    )
    async def refresh_data_source(
        self, data_source_id: str, knowledge_service: KnowledgeService
    ) -> RefreshDataSourceResponse:
        job_id = await knowledge_service.refresh_data_source(data_source_id)
        return RefreshDataSourceResponse(job_id=job_id)

    @get(
        "/collections/{collection_id:str}/search",
        return_dto=SearchResultsResponseDTO,
    )
    async def search_collection(
        self,
        collection_id: str,
        knowledge_service: KnowledgeService,
        search_query: str,
        data_source_ids: list[str] | None = None,
        k: int = 5,
        retrieval_mode: str | None = "dense",
    ) -> SearchCollectionResponse:
        filter = (
            Filter(
                must=[
                    FieldCondition(
                        key="data_source_id",
                        match=MatchAny(any=data_source_ids),
                    ),
                ],
            )
            if data_source_ids
            else None
        )
        chunks = await knowledge_service.search_collection(
            collection_id=collection_id,
            query=search_query,
            filter=filter,
            k=k,
            retrieval_mode=RetrievalMode(retrieval_mode),
        )
        return SearchCollectionResponse(chunks=chunks)



from ..shared.pg import get_engine, get_session_maker


class BaseService:
    def __init__(self, *, user_id: str):
        self.user_id = user_id
        self.db_engine = get_engine()
        self.db_session_maker = get_session_maker()



import json
from datetime import datetime
from typing import Any, AsyncGenerator
from zoneinfo import ZoneInfo

from google.adk.agents.run_config import RunConfig, StreamingMode
from google.adk.runners import Runner
from google.adk.sessions.database_session_service import StorageEvent
from google.genai import types
from litestar.response import ServerSentEventMessage
from sparkden.adk.session_service import SessionService
from sparkden.assistants import assistants
from sparkden.assistants.helpers import assistant_context_collection_id
from sparkden.knowledge.data_loader import DataLoader
from sparkden.models.assistant import Assistant, ContextWithFile
from sparkden.models.knowledge import FileObject, KnowledgeDataSourceCreate
from sparkden.models.shared import OffsetPagination
from sparkden.services.base import BaseService
from sparkden.services.knowledge import KnowledgeService
from sqlalchemy import delete, func, select
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.ext.asyncio import AsyncSession

from ..db.schema import Task as StorageTask
from ..models.task import (
    DataSourceContext,
    InlineContext,
    Task,
    TaskItem,
    TaskSSEData,
    UserContentWithContext,
)


class TaskService(BaseService):
    def __init__(self, user_id: str):
        super().__init__(user_id=user_id)
        self.agent_session_service = SessionService()
        self.knowledge_service = KnowledgeService(user_id=user_id)

    async def list_tasks(
        self, *, limit: int = 30, offset: int = 0
    ) -> OffsetPagination[TaskItem]:
        async with self.db_session_maker.begin() as db_session:
            query = (
                select(
                    StorageTask,
                    func.count().over().label("total_count"),
                )
                .where(StorageTask.user_id == self.user_id)
                .order_by(StorageTask.updated_at.desc())
                .offset(offset)
                .limit(limit)
            )
            result = (await db_session.execute(query)).all()
            tasks = [TaskItem.from_storage(row[0]) for row in result]
            total = result[0][1] if result else 0

        return OffsetPagination(
            items=tasks,
            total=total,
            limit=limit,
            offset=offset,
        )

    async def create_task(
        self, *, title: str, assistant_id: str, task_id: str | None = None
    ) -> Task:
        async with self.db_session_maker.begin() as db_session:
            session = await self.agent_session_service.create_session(
                app_name=assistant_id,
                user_id=self.user_id,
                state={
                    "start_time": datetime.now(
                        tz=ZoneInfo("Asia/Shanghai")
                    ).isoformat(),
                },
                db_session=db_session,
            )

            task_attributes = {
                "user_id": self.user_id,
                "assistant_id": assistant_id,
                "title": title,
                "agent_session_id": session.id,
            }
            if task_id:
                task_attributes["id"] = task_id

            task = await db_session.scalar(
                insert(StorageTask)
                .values(
                    **task_attributes,
                )
                .returning(StorageTask)
            )
            if not task:
                raise Exception("Failed to create task")

        return Task.from_storage(task, session)

    async def get_task(self, *, assistant_id: str, task_id: str) -> Task | None:
        async with self.db_session_maker.begin() as db_session:
            task = await db_session.scalar(
                select(StorageTask).where(
                    StorageTask.user_id == self.user_id,
                    StorageTask.assistant_id == assistant_id,
                    StorageTask.id == task_id,
                )
            )

            if task is None:
                return None

            session = await self.agent_session_service.get_session(
                user_id=self.user_id,
                app_name=assistant_id,
                session_id=task.agent_session_id,
                db_session=db_session,
            )

        return Task.from_storage(task, session) if session else None

    async def run_task(
        self,
        *,
        assistant_id: str,
        task_id: str,
        message_content: list[types.Part],
        state_delta: dict[str, Any] | None = None,
        edit_message_id: str | None = None,
        contexts: list[ContextWithFile] | None = None,
    ) -> AsyncGenerator[ServerSentEventMessage, None]:
        assistant = assistants.get_assistant(assistant_id)
        if assistant is None:
            raise ValueError(f"Assistant {assistant_id} not found")

        async with self.db_session_maker.begin() as db_session:
            task = await db_session.get(StorageTask, task_id)
            if task is None:
                raise ValueError(f"Task {task_id} not found")
            session_id = task.agent_session_id

            # truncate events from the message_id
            if edit_message_id:
                await self._truncate_events_after_id(
                    session_id=session_id,
                    event_id=edit_message_id,
                    db_session=db_session,
                )

        context_content: list[types.Part] = []
        if contexts:
            task_contexts: list[DataSourceContext | InlineContext] = []
            for context in contexts:
                task_context = await self._generate_task_context(assistant, context)
                task_contexts.append(task_context)

            if task_contexts:
                user_content_with_context = UserContentWithContext(
                    contexts=task_contexts,
                    instruction="If context already has content, use content directly, otherwise call the search_context tool with data_source_id to retrieve information.",
                )
                context_content.append(
                    types.Part(
                        text=user_content_with_context.model_dump_json(
                            exclude_none=True,
                            by_alias=True,
                        )
                    )
                )

        new_message = types.Content(
            role="user",
            parts=message_content + context_content,
        )

        runner = Runner(
            app_name=assistant_id,
            agent=assistant.root_agent,
            session_service=self.agent_session_service,
            plugins=[
                *assistant.agent_plugins,
            ],
        )

        accumulated_text = ""
        user_event_yielded = False

        async for event in runner.run_async(
            user_id=self.user_id,
            session_id=session_id,
            new_message=new_message,
            run_config=RunConfig(streaming_mode=StreamingMode.SSE),
            state_delta=state_delta,
        ):
            if not user_event_yielded:
                async with self.db_session_maker.begin() as db_session:
                    # get recent user event
                    user_event = await db_session.scalar(
                        select(StorageEvent)
                        .where(
                            StorageEvent.session_id == session_id,
                            StorageEvent.author == "user",
                        )
                        .order_by(StorageEvent.timestamp.desc())
                        .limit(1)
                    )
                if user_event is not None:
                    user_event_yielded = True
                    yield ServerSentEventMessage(
                        event="message",
                        data=TaskSSEData.from_event(
                            user_event.to_event(),
                        ).model_dump_json(
                            exclude_none=True,
                            by_alias=True,
                        ),
                    )

            event_name, event_data = None, None
            if event.partial:
                event_name = "message/partial"
                accumulated_event = event.model_copy(deep=True)
                if (
                    accumulated_event.content
                    and accumulated_event.content.parts
                    and accumulated_event.content.parts[0].text
                ):
                    accumulated_text += accumulated_event.content.parts[0].text
                    accumulated_event.content.parts[0].text = accumulated_text
                # TODO: support partial function call
                event_data = TaskSSEData.from_event(
                    accumulated_event,
                ).model_dump_json(exclude_none=True, by_alias=True)
            else:
                accumulated_text = ""
                event_name = "message"
                event_data = TaskSSEData.from_event(
                    event,
                ).model_dump_json(
                    exclude_none=True,
                    by_alias=True,
                )
            yield ServerSentEventMessage(
                data=event_data,
                event=event_name,
            )

    async def _truncate_events_after_id(
        self, *, session_id: str, event_id: str, db_session: AsyncSession
    ) -> None:
        """
        Truncate the events after the event_id (ADK session operation only)
        """
        event = await db_session.scalar(
            select(StorageEvent).where(
                StorageEvent.id == event_id, StorageEvent.session_id == session_id
            )
        )
        if event is None:
            return

        deleted_event_ids = await db_session.scalars(
            delete(StorageEvent)
            .where(
                StorageEvent.session_id == session_id,
                StorageEvent.timestamp >= event.timestamp,
            )
            .returning(StorageEvent.id)
        )

        # call on_events_deleted callbacks
        callbacks = assistants.get_callbacks("on_events_deleted")
        for callback in callbacks:
            callback(deleted_event_ids.all(), db_session)

    async def _generate_task_context(
        self, assistant: Assistant, context: ContextWithFile
    ) -> DataSourceContext | InlineContext:
        context_creator = assistant.get_context_creator(context.creator_id)
        if not context_creator:
            raise ValueError(f"Context creator {context.creator_id} not found")

        extra_info = {
            **(context_creator.extra_info or {}),
            **(context.extra_info or {}),
        }
        if context.file:
            extra_info["file"] = FileObject(
                name=context.file.filename,
                content=await context.file.read(),
                content_type=context.file.content_type,
            )

        data_source, _, _ = await self.knowledge_service.add_data_source(
            KnowledgeDataSourceCreate(
                collection_id=assistant_context_collection_id(assistant.id),
                creator_id=context.creator_id,
                name=context.name,
                extra_info=extra_info,
            ),
            auto_load=False,
        )

        data_loader = DataLoader(data_source)
        parsed_result = await data_loader.parse()

        result_str = (
            parsed_result.data
            if isinstance(parsed_result.data, str)
            else json.dumps(parsed_result.data)
        )

        if len(result_str) > 10000:
            return DataSourceContext(
                name=context.name,
                data_source_id=data_source.id,
                creator_id=context.creator_id,
            )
        else:
            return InlineContext(
                creator_id=context.creator_id,
                name=context.name,
                content=result_str,
            )



from typing import Awaitable, Callable, Sequence

from google.adk.tools import FunctionTool, ToolContext

from sparkden.models.knowledge import DataChunk
from sparkden.services.knowledge import KnowledgeService


class SearchKnowledgeTool(FunctionTool):
    def __init__(self, collection_id: str):
        super().__init__(
            func=self.get_tool_func(collection_id),
        )

    @staticmethod
    def get_tool_func(
        collection_id: str,
    ) -> Callable[[str, ToolContext], Awaitable[Sequence[DataChunk]]]:
        async def search_knowledge(
            query: str, tool_context: ToolContext
        ) -> Sequence[DataChunk]:
            """Retrieve information from the knowledge collection.

            Args:
                query: The query to search for.

            Returns:
                A list of data chunks.
            """
            knowledge_service = KnowledgeService(
                user_id=tool_context._invocation_context.user_id
            )
            return await knowledge_service.search_collection(
                collection_id=collection_id,
                query=query,
            )

        return search_knowledge



import os
from io import BytesIO

from minio import Minio, S3Error

from sparkden.models.knowledge import FileObject

from .utils import getenv

_minio_client: Minio | None = None


def get_minio_client() -> Minio:
    global _minio_client
    if _minio_client is None:
        _minio_client = Minio(
            endpoint=getenv("MINIO_ENDPOINT"),
            access_key=getenv("MINIO_ROOT_USER"),
            secret_key=getenv("MINIO_ROOT_PASSWORD"),
            secure=getenv("MINIO_SECURE", "false").lower() == "true",
            region=os.getenv("MINIO_REGION"),
        )
    return _minio_client


def object_exists(bucket_name: str, object_name: str) -> bool:
    try:
        minio_client = get_minio_client()
        minio_client.stat_object(bucket_name, object_name)
        return True
    except S3Error as e:
        if e.code == "NoSuchKey":
            return False
        raise e


def save_objects(bucket_name: str, objects: list[FileObject]) -> None:
    minio_client = get_minio_client()
    for object in objects:
        minio_client.put_object(
            bucket_name=bucket_name,
            object_name=object.name,
            data=BytesIO(object.content),
            length=len(object.content),
            content_type=object.content_type,
        )


def get_object(bucket_name: str, object_name: str) -> FileObject | None:
    response = None
    try:
        minio_client = get_minio_client()
        response = minio_client.get_object(bucket_name, object_name)
        content_type = response.headers["content-type"]
        file = FileObject(
            name=object_name,
            content=response.data,
            content_type=content_type,
        )
    except S3Error as e:
        if e.code == "NoSuchKey":
            return None
        raise e
    finally:
        if response:
            response.close()
            response.release_conn()

    return file


def delete_objects(bucket_name: str, object_names: list[str]) -> None:
    minio_client = get_minio_client()
    for object_name in object_names:
        minio_client.remove_object(
            bucket_name=bucket_name,
            object_name=object_name,
        )



from sparkden.db.schema import DataSourceStatus
from sparkden.knowledge.data_loader import DataLoader
from sparkden.knowledge.registry import knowledge_collections
from sparkden.knowledge.vector_store.base import (
    FieldCondition,
    Filter,
    MatchValue,
)
from sparkden.queued_job.decorator import JobContext, JobResult, queued_job
from sparkden.services.knowledge import KnowledgeService


@queued_job(queue_name="search_context")
async def search_context_job(
    ctx: JobContext,
    data_source_id: str,
    query: str,
) -> JobResult:
    """
    Asynchronously processes a data source, including loading, parsing,
    embedding, and indexing.
    """
    knowledge_service = KnowledgeService(user_id=ctx.user_id)

    try:
        await ctx.update_progress(5, custom_status="initializing")

        data_source = await knowledge_service.get_data_source(data_source_id)
        if not data_source:
            raise ValueError(f"Data source {data_source_id} not found")

        if data_source.status != DataSourceStatus.READY:
            collection = knowledge_collections.get_collection(data_source.collection_id)
            if not collection:
                raise ValueError(f"Collection {data_source.collection_id} not found")

            loader = DataLoader(
                data_source=data_source,
                vector_dimensions=collection.vector_dimensions,
                retrieval_mode=collection.retrieval_mode,
            )

            async for status in loader.load():
                await ctx.update_progress(30, custom_status=status)

            await knowledge_service.update_data_source(
                data_source_id, {"status": DataSourceStatus.READY}
            )

        await ctx.update_progress(90, custom_status="retrieving")

        filter = Filter(
            must=[
                FieldCondition(
                    key="data_source_id",
                    match=MatchValue(value=data_source_id),
                ),
            ],
        )
        data_chunks = await knowledge_service.search_collection(
            collection_id=data_source.collection_id,
            query=query,
            filter=filter,
        )

        await ctx.update_progress(100, custom_status="completed")

        return {
            "data_chunks": [chunk.model_dump() for chunk in data_chunks],
            "data_source_creator_id": data_source.creator_id,
            "data_source_name": data_source.name,
        }

    except Exception as e:
        # Update data source status to ERROR on failure
        await knowledge_service.update_data_source(
            data_source_id, {"status": DataSourceStatus.ERROR}
        )
        # Re-raise the exception so the job is marked as failed
        raise e
